import warnings
from typing import Tuple

import numpy as np
import pandas as pd

from bofire.utils.tmpfile import make_tmpfile

try:
    from xgboost import XGBRegressor  # type: ignore
except ImportError:
    warnings.warn("xgboost not installed, BoFire's `XGBoostSurrogate` cannot be used.")

import uuid

from bofire.data_models.surrogates.api import XGBoostSurrogate as DataModel
from bofire.surrogates.surrogate import Surrogate
from bofire.surrogates.trainable import TrainableSurrogate


class XGBoostSurrogate(TrainableSurrogate, Surrogate):
    def __init__(self, data_model: DataModel, **kwargs) -> None:
        self.n_estimators = data_model.n_estimators
        self.max_depth = data_model.max_depth
        self.max_leaves = data_model.max_leaves
        self.max_bin = data_model.max_bin
        self.grow_policy = data_model.grow_policy
        self.learning_rate = data_model.learning_rate
        self.objective = data_model.objective
        self.booster = data_model.booster
        self.n_jobs = data_model.n_jobs
        self.gamma = data_model.gamma
        self.min_child_weight = data_model.min_child_weight
        self.max_delta_step = data_model.max_delta_step
        self.subsample = data_model.subsample
        self.sampling_method = data_model.sampling_method
        self.colsample_bytree = data_model.colsample_bytree
        self.colsample_bylevel = data_model.colsample_bylevel
        self.colsample_bynode = data_model.colsample_bynode
        self.reg_alpha = data_model.reg_alpha
        self.reg_lambda = data_model.reg_lambda
        self.scale_pos_weight = data_model.scale_pos_weight
        self.random_state = data_model.random_state
        self.num_parallel_tree = data_model.num_parallel_tree
        self.tmpfile_name = f"xgb_{uuid.uuid4().hex}.json"
        super().__init__(data_model=data_model, **kwargs)

    def _init_xgb(self):
        self.model = XGBRegressor(
            n_estimators=self.n_estimators,
            max_depth=self.max_depth,
            max_leaves=self.max_leaves,
            max_bin=self.max_bin,
            grow_policy=self.grow_policy,
            learning_rate=self.learning_rate,
            objective=self.objective,
            booster=self.booster,
            n_jobs=self.n_jobs,
            gamma=self.gamma,
            min_child_weight=self.min_child_weight,
            max_delta_step=self.max_delta_step,
            subsample=self.subsample,
            sampling_method=self.sampling_method,
            colsample_bytree=self.colsample_bytree,
            colsample_bylevel=self.colsample_bylevel,
            colsample_bynode=self.colsample_bynode,
            reg_alpha=self.reg_alpha,
            reg_lambda=self.reg_lambda,
            scale_pos_weight=self.scale_pos_weight,
            random_state=self.random_state,
            num_parallel_tree=self.num_parallel_tree,
        )

    def _fit(self, X: pd.DataFrame, Y: pd.DataFrame, **kwargs):
        transformed_X = self.inputs.transform(X, self.input_preprocessing_specs)
        self._init_xgb()
        self.model.fit(X=transformed_X.values, y=Y.values)

    def _predict(self, transformed_X: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        preds = self.model.predict(transformed_X.values)
        return preds.reshape((transformed_X.shape[0], 1)), np.zeros(
            (transformed_X.shape[0], 1)
        )

    def loads(self, data: str):
        with make_tmpfile(name=self.tmpfile_name) as fname:
            # write to file
            self._init_xgb()
            with open(fname, "w") as f:
                f.write(data)
            self.model.load_model(fname)

    def _dumps(self) -> str:
        with make_tmpfile(name=self.tmpfile_name) as fname:
            self.model.save_model(fname=fname)
            with open(fname, "r") as f:
                dump = f.read()
            return dump



from typing import Literal, Union

import numpy as np
import pandas as pd

from bofire.data_models.objectives.objective import (
    ConstrainedObjective,
    Objective,
    TGe0,
    TGt0,
    TWeight,
)


class CloseToTargetObjective(Objective):
    """Optimize towards a target value. It can be used as objective
    in multiobjective scenarios.

    Attributes:
        w (float): float between zero and one for weighting the objective.
        target_value (float): target value that should be reached.
        exponent (float): the exponent of the expression.
    """

    type: Literal["CloseToTargetObjective"] = "CloseToTargetObjective"
    w: TWeight = 1
    target_value: float
    exponent: float

    def __call__(self, x: Union[pd.Series, np.ndarray]) -> Union[pd.Series, np.ndarray]:
        return -1 * (np.abs(x - self.target_value) ** self.exponent)


class TargetObjective(Objective, ConstrainedObjective):
    """Class for objectives for optimizing towards a target value

    Attributes:
        w (float): float between zero and one for weighting the objective.
        target_value (float): target value that should be reached.
        tolerance (float): Tolerance for reaching the target. Has to be greater than zero.
        steepness (float): Steepness of the sigmoid function. Has to be greater than zero.

    """

    type: Literal["TargetObjective"] = "TargetObjective"
    w: TWeight = 1
    target_value: float
    tolerance: TGe0
    steepness: TGt0

    def __call__(self, x: Union[pd.Series, np.ndarray]) -> Union[pd.Series, np.ndarray]:
        """The call function returning a reward for passed x values.

        Args:
            x (np.array): An array of x values

        Returns:
            np.array: An array of reward values calculated by the product of two sigmoidal shaped functions resulting in a maximum at the target value.
        """
        return (
            1
            / (
                1
                + np.exp(
                    -1 * self.steepness * (x - (self.target_value - self.tolerance))
                )
            )
            * (
                1
                - 1
                / (
                    1.0
                    + np.exp(
                        -1 * self.steepness * (x - (self.target_value + self.tolerance))
                    )
                )
            )
        )



from pydantic import parse_obj_as

from bofire.data_models.api import (
    AnyAcquisitionFunction,
    AnyCondition,
    AnyConstraint,
    AnyFeature,
    AnyKernel,
    AnyMolFeatures,
    AnyObjective,
    AnyOutlierDetection,
    AnyPrior,
    AnyStrategy,
    AnySurrogate,
    Domain,
)
from tests.bofire.data_models.specs.api import Spec


def test_prior_should_be_deserializable(prior_spec: Spec):
    obj = prior_spec.obj()
    deserialized = parse_obj_as(AnyPrior, obj.dict())
    assert obj == deserialized


def test_kernel_should_be_deserializable(kernel_spec: Spec):
    obj = kernel_spec.obj()
    deserialized = parse_obj_as(AnyKernel, obj.dict())
    assert obj == deserialized


def test_constraint_should_be_deserializable(constraint_spec: Spec):
    obj = constraint_spec.obj()
    deserialized = parse_obj_as(AnyConstraint, obj.dict())
    assert obj == deserialized


def test_objective_should_be_deserializable(objective_spec: Spec):
    obj = objective_spec.obj()
    deserialized = parse_obj_as(AnyObjective, obj.dict())
    assert obj == deserialized


def test_feature_should_be_deserializable(feature_spec: Spec):
    obj = feature_spec.obj()
    deserialized = parse_obj_as(AnyFeature, obj.dict())
    assert obj == deserialized


def test_domain_should_be_deserializable(domain_spec: Spec):
    obj = domain_spec.obj()
    deserialized = parse_obj_as(Domain, obj.dict())
    assert obj == deserialized


def test_surrogate_should_be_deserializable(surrogate_spec: Spec):
    obj = surrogate_spec.obj()
    deserialized = parse_obj_as(AnySurrogate, obj.dict())
    assert obj == deserialized


def test_acquisition_function_should_be_deserializable(acquisition_function_spec: Spec):
    obj = acquisition_function_spec.obj()
    deserialized = parse_obj_as(AnyAcquisitionFunction, obj.dict())
    assert obj == deserialized


def test_strategy_should_be_deserializable(strategy_spec: Spec):
    obj = strategy_spec.obj()
    deserialized = parse_obj_as(AnyStrategy, obj.dict())
    # TODO: can we unhide the comparison of surrogate_specs?
    obj = {k: v for k, v in obj.dict().items() if k != "surrogate_specs"}
    deserialized = {
        k: v for k, v in deserialized.dict().items() if k != "surrogate_specs"
    }
    assert obj == deserialized


def test_condition_should_be_deserializable(condition_spec: Spec):
    obj = condition_spec.obj()
    deserialized = parse_obj_as(AnyCondition, obj.dict())
    assert obj == deserialized


def test_outlier_detection_should_be_deserializable(outlier_detection_spec: Spec):
    obj = outlier_detection_spec.obj()
    deserialized = parse_obj_as(AnyOutlierDetection, obj.dict())
    assert obj == deserialized


def test_molfeatures_should_be_deserializable(molfeatures_spec: Spec):
    obj = molfeatures_spec.obj()
    deserialized = parse_obj_as(AnyMolFeatures, obj.dict())
    assert obj == deserialized



import numpy as np
from pandas.testing import assert_frame_equal

import bofire.benchmarks.benchmark as benchmark
import bofire.strategies.api as strategies
from bofire.benchmarks.api import GenericBenchmark
from bofire.benchmarks.multi import ZDT1
from bofire.benchmarks.single import Himmelblau
from bofire.data_models.strategies.api import PolytopeSampler


def test_generic_benchmark():
    bench = ZDT1(n_inputs=5)
    genbench = GenericBenchmark(domain=bench.domain, func=bench._f)
    candidates = bench.domain.inputs.sample(5)
    samples1 = bench.f(candidates=candidates)
    samples2 = genbench.f(candidates=candidates)
    assert_frame_equal(samples1, samples2)


def test_benchmark_generate_outliers():
    def sample(domain):
        datamodel = PolytopeSampler(domain=domain)
        sampler = strategies.map(data_model=datamodel)
        sampled = sampler.ask(10)
        return sampled

    outlier_rate = 0.5
    Benchmark = Himmelblau()
    sampled = sample(Benchmark.domain)
    sampled_xy = Benchmark.f(sampled, return_complete=True)
    Benchmark = Himmelblau(
        outlier_rate=outlier_rate,
        outlier_prior=benchmark.UniformOutlierPrior(bounds=(50, 100)),
    )
    sampled_xy1 = Benchmark.f(sampled, return_complete=True)
    assert np.sum(sampled_xy["y"] != sampled_xy1["y"]) != 0
    assert isinstance(Benchmark.outlier_prior, benchmark.UniformOutlierPrior)



import random

import numpy as np
import pytest
import torch
from botorch.acquisition.objective import ConstrainedMCObjective, GenericMCObjective
from botorch.utils.objective import soft_eval_constraint

import bofire.strategies.api as strategies
from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.domain.api import Constraints, Domain, Inputs, Outputs
from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.objectives.api import (
    CloseToTargetObjective,
    MaximizeObjective,
    MaximizeSigmoidObjective,
    MinimizeObjective,
    MinimizeSigmoidObjective,
    TargetObjective,
)
from bofire.data_models.strategies.api import PolytopeSampler
from bofire.utils.torch_tools import (
    constrained_objective2botorch,
    get_additive_botorch_objective,
    get_custom_botorch_objective,
    get_initial_conditions_generator,
    get_linear_constraints,
    get_multiobjective_objective,
    get_multiplicative_botorch_objective,
    get_nchoosek_constraints,
    get_objective_callable,
    get_output_constraints,
    tkwargs,
)

if1 = ContinuousInput(
    bounds=(0, 1),
    key="if1",
)
if2 = ContinuousInput(
    bounds=(0, 1),
    key="if2",
)
if3 = ContinuousInput(
    bounds=(0, 1),
    key="if3",
)
if4 = ContinuousInput(
    bounds=(0.1, 0.1),
    key="if4",
)
if5 = CategoricalInput(
    categories=["a", "b", "c"],
    key="if5",
)
if6 = CategoricalInput(
    categories=["a", "b", "c"],
    allowed=[False, True, False],
    key="if6",
)
c1 = LinearEqualityConstraint(
    features=["if1", "if2", "if3", "if4"], coefficients=[1.0, 1.0, 1.0, 1.0], rhs=1.0
)
c2 = LinearInequalityConstraint(
    features=["if1", "if2"], coefficients=[1.0, 1.0], rhs=0.2
)
c3 = LinearInequalityConstraint(
    features=["if1", "if2", "if4"], coefficients=[1.0, 1.0, 0.5], rhs=0.2
)


@pytest.mark.parametrize(
    "objective",
    [
        MaximizeObjective(w=0.5),
        MaximizeSigmoidObjective(steepness=1.0, tp=1.0, w=0.5),
        MinimizeObjective(w=0.5),
        MinimizeSigmoidObjective(steepness=1.0, tp=1.0, w=0.5),
        TargetObjective(target_value=2.0, steepness=1.0, tolerance=1e-3, w=0.5),
        CloseToTargetObjective(target_value=2.0, exponent=1.0, w=0.5),
        # ConstantObjective(w=0.5, value=1.0),
    ],
)
def test_get_objective_callable(objective):
    samples = (torch.rand(50, 3, requires_grad=True) * 5.0).to(**tkwargs)
    a_samples = samples.detach().numpy()
    callable = get_objective_callable(idx=1, objective=objective)
    assert np.allclose(
        # objective.reward(samples, desFunc)[0].detach().numpy(),
        callable(samples).detach().numpy(),
        objective(a_samples[:, 1]),
        rtol=1e-06,
    )


def f1(samples, callables, weights, X):
    outputs_list = []
    for c, w in zip(callables, weights):
        outputs_list.append(c(samples, None) ** w)
    samples = torch.stack(outputs_list, dim=-1)

    return (samples[..., 0] + samples[..., 1]) * (samples[..., 0] * samples[..., 1])


def f2(samples, callables, weights, X):
    outputs_list = []
    for c, w in zip(callables, weights):
        outputs_list.append(c(samples, None) ** w)
    samples = torch.stack(outputs_list, dim=-1)

    return (
        (samples[..., 0] + samples[..., 1])
        * (samples[..., 0] * samples[..., 1])
        * (samples[..., 0] * samples[..., 2])
    )


@pytest.mark.parametrize("f, exclude_constraints", [(f1, True), (f2, False)])
def test_get_custom_botorch_objective(f, exclude_constraints):
    samples = (torch.rand(30, 3, requires_grad=True) * 5).to(**tkwargs)
    a_samples = samples.detach().numpy()
    obj1 = MaximizeObjective(w=1.0)
    obj2 = MinimizeSigmoidObjective(steepness=1.0, tp=1.0, w=1.0)
    obj3 = CloseToTargetObjective(w=1.0, target_value=2, exponent=1)
    outputs = Outputs(
        features=[
            ContinuousOutput(
                key="alpha",
                objective=obj1,
            ),
            ContinuousOutput(
                key="beta",
                objective=obj2,
            ),
            ContinuousOutput(
                key="gamma",
                objective=obj3,
            ),
        ]
    )
    objective = get_custom_botorch_objective(
        outputs, f=f, exclude_constraints=exclude_constraints
    )
    generic_objective = GenericMCObjective(objective=objective)
    objective_forward = generic_objective.forward(samples)

    # calc with numpy
    reward1 = obj1(a_samples[:, 0])
    reward2 = obj2(a_samples[:, 1])
    reward3 = obj3(a_samples[:, 2])
    # do the comparison
    assert np.allclose(
        (reward1**obj1.w + reward3**obj3.w)
        * (reward1**obj1.w * reward3**obj3.w)
        if exclude_constraints
        else (reward1**obj1.w + reward2**obj2.w)
        * (reward1**obj1.w * reward2**obj2.w)
        * (reward1**obj1.w * reward3**obj3.w),
        objective_forward.detach().numpy(),
        rtol=1e-06,
    )
    if exclude_constraints:
        constraints, etas = get_output_constraints(outputs=outputs)
        generic_objective = ConstrainedMCObjective(
            objective=objective,
            constraints=constraints,
            eta=torch.tensor(etas).to(**tkwargs),
        )
        objective_forward = generic_objective.forward(samples)
        assert np.allclose(
            np.clip(
                (
                    (reward1**obj1.w + reward3**obj3.w)
                    * (reward1**obj1.w * reward3**obj3.w)
                )
                * reward2,
                0,
                None,
            ),
            objective_forward.detach().numpy(),
            rtol=1e-06,
        )


def test_get_multiplicative_botorch_objective():
    (obj1, obj2) = random.choices(
        [
            MaximizeObjective(w=0.5),
            MaximizeSigmoidObjective(steepness=1.0, tp=1.0, w=0.5),
            MinimizeObjective(w=1),
            MinimizeSigmoidObjective(steepness=1.0, tp=1.0, w=0.5),
            TargetObjective(target_value=2.0, steepness=1.0, tolerance=1e-3, w=0.5),
            CloseToTargetObjective(target_value=2.0, exponent=1.0, w=1.0),
        ],
        k=2,
    )
    outputs = Outputs(
        features=[
            ContinuousOutput(key="alpha", objective=obj1),
            ContinuousOutput(key="beta", objective=obj2),
        ]
    )
    objective = get_multiplicative_botorch_objective(outputs)
    generic_objective = GenericMCObjective(objective=objective)
    samples = (torch.rand(30, 2, requires_grad=True) * 5).to(**tkwargs)
    a_samples = samples.detach().numpy()
    objective_forward = generic_objective.forward(samples)
    # calc with numpy
    reward1 = obj1(a_samples[:, 0])
    reward2 = obj2(a_samples[:, 1])
    # do the comparison
    assert np.allclose(
        # objective.reward(samples, desFunc)[0].detach().numpy(),
        reward1**obj1.w * reward2**obj2.w,
        objective_forward.detach().numpy(),
        rtol=1e-06,
    )


@pytest.mark.parametrize("exclude_constraints", [True, False])
def test_get_additive_botorch_objective(exclude_constraints):
    samples = (torch.rand(30, 3, requires_grad=True) * 5).to(**tkwargs)
    a_samples = samples.detach().numpy()
    obj1 = MaximizeObjective(w=0.5)
    obj2 = MinimizeSigmoidObjective(steepness=1.0, tp=1.0, w=0.5)
    obj3 = CloseToTargetObjective(w=0.5, target_value=2, exponent=1)
    # obj3 = MaximizeObjective(w=0.7)
    outputs = Outputs(
        features=[
            ContinuousOutput(
                key="alpha",
                objective=obj1,
            ),
            ContinuousOutput(
                key="beta",
                objective=obj2,
            ),
            ContinuousOutput(
                key="gamma",
                objective=obj3,
            ),
        ]
    )
    objective = get_additive_botorch_objective(
        outputs, exclude_constraints=exclude_constraints
    )
    generic_objective = GenericMCObjective(objective=objective)
    objective_forward = generic_objective.forward(samples)

    # calc with numpy
    reward1 = obj1(a_samples[:, 0])
    reward2 = obj2(a_samples[:, 1])
    reward3 = obj3(a_samples[:, 2])
    # do the comparison
    assert np.allclose(
        # objective.reward(samples, desFunc)[0].detach().numpy(),
        reward1 * obj1.w + reward3 * obj3.w
        if exclude_constraints
        else reward1 * obj1.w + reward3 * obj3.w + reward2 * obj2.w,
        objective_forward.detach().numpy(),
        rtol=1e-06,
    )
    if exclude_constraints:
        constraints, etas = get_output_constraints(outputs=outputs)
        generic_objective = ConstrainedMCObjective(
            objective=objective,
            constraints=constraints,
            eta=torch.tensor(etas).to(**tkwargs),
        )
        objective_forward = generic_objective.forward(samples)
        assert np.allclose(
            np.clip((reward1 * obj1.w + reward3 * obj3.w) * reward2, 0, None),
            objective_forward.detach().numpy(),
            rtol=1e-06,
        )


def test_get_linear_constraints():
    domain = Domain(inputs=[if1, if2])
    constraints = get_linear_constraints(domain, LinearEqualityConstraint)
    assert len(constraints) == 0
    constraints = get_linear_constraints(domain, LinearInequalityConstraint)
    assert len(constraints) == 0

    domain = Domain(inputs=[if1, if2, if3], constraints=[c2])
    constraints = get_linear_constraints(domain, LinearEqualityConstraint)
    assert len(constraints) == 0
    constraints = get_linear_constraints(domain, LinearInequalityConstraint)
    assert len(constraints) == 1
    assert constraints[0][2] == c2.rhs * -1
    assert torch.allclose(constraints[0][0], torch.tensor([0, 1]))
    assert torch.allclose(constraints[0][1], torch.tensor([-1.0, -1.0]).to(**tkwargs))

    domain = Domain(
        inputs=[if1, if2, if3, if4],
        constraints=[c1, c2],
    )
    constraints = get_linear_constraints(domain, LinearEqualityConstraint)
    assert len(constraints) == 1
    assert constraints[0][2] == -1 * (c1.rhs - 0.1)
    assert len(constraints[0][0]) == len(c1.features) - 1
    assert torch.allclose(constraints[0][0], torch.tensor([0, 1, 2]))
    assert torch.allclose(constraints[0][1], torch.tensor([-1, -1, -1]).to(**tkwargs))
    assert len(constraints[0][1]) == len(c1.coefficients) - 1
    constraints = get_linear_constraints(domain, LinearInequalityConstraint)
    assert len(constraints) == 1
    assert constraints[0][2] == c2.rhs * -1
    assert len(constraints[0][0]) == len(c2.features)
    assert len(constraints[0][1]) == len(c2.coefficients)
    assert torch.allclose(constraints[0][0], torch.tensor([0, 1]))
    assert torch.allclose(constraints[0][1], torch.tensor([-1.0, -1.0]).to(**tkwargs))

    domain = Domain(
        inputs=[if1, if2, if3, if4, if5],
        constraints=[c1, c2, c3],
    )
    constraints = get_linear_constraints(domain, LinearEqualityConstraint)
    assert len(constraints) == 1
    assert constraints[0][2] == (c1.rhs - 0.1) * -1
    assert len(constraints[0][0]) == len(c1.features) - 1
    assert len(constraints[0][1]) == len(c1.coefficients) - 1
    assert torch.allclose(constraints[0][0], torch.tensor([0, 1, 2]))
    assert torch.allclose(constraints[0][1], torch.tensor([-1, -1, -1]).to(**tkwargs))
    constraints = get_linear_constraints(domain, LinearInequalityConstraint)
    assert len(constraints) == 2
    assert constraints[0][2] == c2.rhs * -1
    assert len(constraints[0][0]) == len(c2.features)
    assert len(constraints[0][1]) == len(c2.coefficients)
    assert torch.allclose(constraints[0][0], torch.tensor([0, 1]))
    assert torch.allclose(constraints[0][1], torch.tensor([-1.0, -1.0]).to(**tkwargs))
    assert constraints[1][2] == (c3.rhs - 0.5 * 0.1) * -1
    assert len(constraints[1][0]) == len(c3.features) - 1
    assert len(constraints[1][1]) == len(c3.coefficients) - 1
    assert torch.allclose(constraints[1][0], torch.tensor([0, 1]))
    assert torch.allclose(constraints[1][1], torch.tensor([-1.0, -1.0]).to(**tkwargs))


def test_get_linear_constraints_unit_scaled():
    inputs = [
        ContinuousInput(
            key="base_polymer",
            bounds=(0.3, 0.7),
        ),
        ContinuousInput(
            key="glas_fibre",
            bounds=(0.1, 0.7),
        ),
        ContinuousInput(
            key="additive",
            bounds=(0.1, 0.6),
        ),
        ContinuousInput(
            key="temperature",
            bounds=(30, 700),
        ),
    ]
    constraints = [
        LinearEqualityConstraint(
            coefficients=[1.0, 1.0, 1.0],
            features=["base_polymer", "glas_fibre", "additive"],
            rhs=1.0,
        )
    ]
    domain = Domain(inputs=inputs, constraints=constraints)

    constraints = get_linear_constraints(
        domain, LinearEqualityConstraint, unit_scaled=True
    )
    assert len(constraints) == 1
    assert len(constraints[0][0]) == 3
    assert len(constraints[0][1]) == 3
    assert constraints[0][2] == 0.5 * -1
    assert torch.allclose(
        constraints[0][1], torch.tensor([0.4, 0.6, 0.5]).to(**tkwargs) * -1
    )
    assert torch.allclose(constraints[0][0], torch.tensor([1, 2, 0]))


of1 = ContinuousOutput(key="of1", objective=MaximizeObjective(w=1.0))
of2 = ContinuousOutput(
    key="of2",
    objective=MaximizeSigmoidObjective(
        w=1.0,
        tp=0,
        steepness=2,
    ),
)
of3 = ContinuousOutput(
    key="of3",
    objective=TargetObjective(w=1.0, tolerance=2, target_value=5, steepness=4),
)


@pytest.mark.parametrize(
    "outputs",
    [
        Outputs(features=[of1, of2, of3]),
        Outputs(features=[of2, of1, of3]),
    ],
)
def test_get_output_constraints(outputs):
    constraints, etas = get_output_constraints(outputs=outputs)
    assert len(constraints) == len(etas)
    assert np.allclose(etas, [0.5, 0.25, 0.25])


def test_get_nchoosek_constraints():
    domain = Domain(
        inputs=Inputs(
            features=[ContinuousInput(key=f"if{i+1}", bounds=(0, 1)) for i in range(8)]
        ),
        constraints=Constraints(
            constraints=[
                NChooseKConstraint(
                    features=[f"if{i+3}" for i in range(6)],
                    min_count=2,
                    max_count=5,
                    none_also_valid=False,
                )
            ]
        ),
    )
    constraints = get_nchoosek_constraints(domain=domain)
    assert len(constraints) == 2
    # wrong samples
    samples = domain.inputs.sample(5)
    # check max count not fulfilled
    assert torch.all(constraints[0](torch.from_numpy(samples.values).to(**tkwargs)) < 0)
    # check max count fulfilled
    samples.if8 = 0
    assert torch.all(
        constraints[0](torch.from_numpy(samples.values).to(**tkwargs)) >= 0
    )

    # check min count fulfilled
    samples = domain.inputs.sample(5)
    assert torch.all(
        constraints[1](torch.from_numpy(samples.values).to(**tkwargs)) >= 0
    )
    samples[[f"if{i+4}" for i in range(5)]] = 0.0
    assert torch.all(constraints[1](torch.from_numpy(samples.values).to(**tkwargs)) < 0)
    domain = Domain(
        inputs=Inputs(
            features=[ContinuousInput(key=f"if{i+1}", bounds=(0, 1)) for i in range(8)]
        ),
        constraints=Constraints(
            constraints=[
                NChooseKConstraint(
                    features=[f"if{i+3}" for i in range(6)],
                    min_count=3,
                    max_count=6,
                    none_also_valid=False,
                )
            ]
        ),
    )
    constraints = get_nchoosek_constraints(domain=domain)
    assert len(constraints) == 1
    samples = domain.inputs.sample(5)
    assert torch.all(
        constraints[0](torch.from_numpy(samples.values).to(**tkwargs)) >= 0
    )

    domain = Domain(
        inputs=Inputs(
            features=[ContinuousInput(key=f"if{i+1}", bounds=(0, 1)) for i in range(8)]
        ),
        constraints=Constraints(
            constraints=[
                NChooseKConstraint(
                    features=[f"if{i+3}" for i in range(6)],
                    min_count=0,
                    max_count=2,
                    none_also_valid=False,
                )
            ]
        ),
    )
    constraints = get_nchoosek_constraints(domain=domain)
    assert len(constraints) == 1
    samples = domain.inputs.sample(5)
    assert torch.all(constraints[0](torch.from_numpy(samples.values).to(**tkwargs)) < 0)
    # test with two max nchoosek constraints
    domain = Domain(
        inputs=[
            ContinuousInput(key="x1", bounds=[0, 1]),
            ContinuousInput(key="x2", bounds=[0, 1]),
            ContinuousInput(key="x3", bounds=[0, 1]),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NChooseKConstraint(
                features=["x1", "x2", "x3"],
                min_count=0,
                max_count=1,
                none_also_valid=False,
            ),
            NChooseKConstraint(
                features=["x1", "x2", "x3"],
                min_count=0,
                max_count=2,
                none_also_valid=False,
            ),
        ],
    )
    samples = torch.tensor([[1, 0, 0], [1, 1, 0], [1, 1, 1]]).to(**tkwargs)
    constraints = get_nchoosek_constraints(domain=domain)
    assert torch.allclose(
        constraints[0](samples), torch.tensor([0.0, -1.0, -2.0]).to(**tkwargs)
    )
    assert torch.allclose(
        constraints[1](samples), torch.tensor([1.0, 0.0, -1.0]).to(**tkwargs)
    )
    # test with two min nchoosek constraints
    domain = Domain(
        inputs=[
            ContinuousInput(key="x1", bounds=[0, 1]),
            ContinuousInput(key="x2", bounds=[0, 1]),
            ContinuousInput(key="x3", bounds=[0, 1]),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NChooseKConstraint(
                features=["x1", "x2", "x3"],
                min_count=1,
                max_count=3,
                none_also_valid=False,
            ),
            NChooseKConstraint(
                features=["x1", "x2", "x3"],
                min_count=2,
                max_count=3,
                none_also_valid=False,
            ),
        ],
    )
    samples = torch.tensor([[1, 0, 0], [1, 1, 0], [1, 1, 1]]).to(**tkwargs)
    constraints = get_nchoosek_constraints(domain=domain)
    assert torch.allclose(
        constraints[0](samples), torch.tensor([0.0, 1.0, 2.0]).to(**tkwargs)
    )
    assert torch.allclose(
        constraints[1](samples), torch.tensor([-1.0, 0.0, 1.0]).to(**tkwargs)
    )
    # test with min/max and max constraint
    # test with two min nchoosek constraints
    domain = Domain(
        inputs=[
            ContinuousInput(key="x1", bounds=[0, 1]),
            ContinuousInput(key="x2", bounds=[0, 1]),
            ContinuousInput(key="x3", bounds=[0, 1]),
        ],
        outputs=[ContinuousOutput(key="y")],
        constraints=[
            NChooseKConstraint(
                features=["x1", "x2", "x3"],
                min_count=1,
                max_count=2,
                none_also_valid=False,
            ),
            NChooseKConstraint(
                features=["x1", "x2", "x3"],
                min_count=0,
                max_count=2,
                none_also_valid=False,
            ),
        ],
    )
    samples = torch.tensor([[1, 0, 0], [1, 1, 0], [1, 1, 1]]).to(**tkwargs)
    constraints = get_nchoosek_constraints(domain=domain)
    assert torch.allclose(
        constraints[0](samples), torch.tensor([1.0, 0.0, -1.0]).to(**tkwargs)
    )
    assert torch.allclose(
        constraints[1](samples), torch.tensor([0.0, 1.0, 2.0]).to(**tkwargs)
    )
    assert torch.allclose(
        constraints[2](samples), torch.tensor([1.0, 0.0, -1.0]).to(**tkwargs)
    )


def test_get_multiobjective_objective():
    samples = (torch.rand(30, 4, requires_grad=True) * 5).to(**tkwargs)
    samples2 = (torch.rand(30, 512, 4, requires_grad=True) * 5).to(**tkwargs)
    a_samples = samples.detach().numpy()
    obj1 = MaximizeObjective()
    obj2 = MinimizeSigmoidObjective(steepness=1.0, tp=1.0, w=0.5)
    obj3 = MinimizeObjective()
    obj4 = CloseToTargetObjective(target_value=2.5, exponent=1)
    outputs = Outputs(
        features=[
            ContinuousOutput(
                key="alpha",
                objective=obj1,
            ),
            ContinuousOutput(
                key="beta",
                objective=obj2,
            ),
            ContinuousOutput(
                key="gamma",
                objective=obj3,
            ),
            ContinuousOutput(
                key="omega",
                objective=obj4,
            ),
        ]
    )
    objective = get_multiobjective_objective(outputs=outputs)
    generic_objective = GenericMCObjective(objective=objective)
    # check the shape
    objective_forward = generic_objective.forward(samples2)
    assert objective_forward.shape == torch.Size((30, 512, 3))
    objective_forward = generic_objective.forward(samples)
    assert objective_forward.shape == torch.Size((30, 3))
    # check what is in
    # calc with numpy
    reward1 = obj1(a_samples[:, 0])
    reward3 = obj3(a_samples[:, 2])
    reward4 = obj4(a_samples[:, 3])
    assert np.allclose(objective_forward[..., 0].detach().numpy(), reward1)
    assert np.allclose(objective_forward[..., 1].detach().numpy(), reward3)
    assert np.allclose(objective_forward[..., 2].detach().numpy(), reward4)


@pytest.mark.parametrize("sequential", [True, False])
def test_get_initial_conditions_generator(sequential: bool):
    inputs = Inputs(
        features=[
            ContinuousInput(key="a", bounds=(0, 1)),
            CategoricalDescriptorInput(
                key="b",
                categories=["alpha", "beta", "gamma"],
                descriptors=["omega"],
                values=[[0], [1], [3]],
            ),
        ]
    )
    domain = Domain(inputs=inputs)
    strategy = strategies.map(PolytopeSampler(domain=domain))
    # test with one hot encoding
    generator = get_initial_conditions_generator(
        strategy=strategy,
        transform_specs={"b": CategoricalEncodingEnum.ONE_HOT},
        ask_options={},
        sequential=sequential,
    )
    initial_conditions = generator(n=3, q=2, seed=42)
    assert initial_conditions.shape == torch.Size((3, 2, 4))
    # test with descriptor encoding
    generator = get_initial_conditions_generator(
        strategy=strategy,
        transform_specs={"b": CategoricalEncodingEnum.DESCRIPTOR},
        ask_options={},
        sequential=sequential,
    )
    initial_conditions = generator(n=3, q=2, seed=42)
    assert initial_conditions.shape == torch.Size((3, 2, 2))


@pytest.mark.parametrize(
    "objective",
    [
        (MaximizeSigmoidObjective(w=1, tp=15, steepness=0.5)),
        (MinimizeSigmoidObjective(w=1, tp=15, steepness=0.5)),
        (TargetObjective(w=1, target_value=15, steepness=2, tolerance=5)),
    ],
)
def test_constrained_objective2botorch(objective):
    cs, etas = constrained_objective2botorch(idx=0, objective=objective)

    x = torch.from_numpy(np.linspace(0, 30, 500)).unsqueeze(-1)
    y = torch.ones([500])

    for c, eta in zip(cs, etas):
        xtt = c(x)
        y *= soft_eval_constraint(xtt, eta)

    assert np.allclose(objective.__call__(np.linspace(0, 30, 500)), y.numpy().ravel())



from typing import ClassVar, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
from pydantic import root_validator, validator

from bofire.data_models.enum import CategoricalEncodingEnum
from bofire.data_models.features.categorical import CategoricalInput
from bofire.data_models.features.continuous import ContinuousInput
from bofire.data_models.features.feature import (
    _CAT_SEP,
    TCategoricalDescriptorVals,
    TDescriptors,
    TDiscreteVals,
    TTransform,
)


# TODO: write a Descriptor base class from which both Categorical and Continuous Descriptor are inheriting
class ContinuousDescriptorInput(ContinuousInput):
    """Class for continuous input features with descriptors

    Attributes:
        lower_bound (float): Lower bound of the feature in the optimization.
        upper_bound (float): Upper bound of the feature in the optimization.
        descriptors (List[str]): Names of the descriptors.
        values (List[float]): Values of the descriptors.
    """

    type: Literal["ContinuousDescriptorInput"] = "ContinuousDescriptorInput"
    order_id: ClassVar[int] = 2

    descriptors: TDescriptors
    values: TDiscreteVals

    @validator("descriptors")
    def descriptors_to_keys(cls, descriptors):
        """validates the descriptor names and transforms it to valid keys

        Args:
            descriptors (List[str]): List of descriptor names

        Returns:
            List[str]: List of valid keys
        """
        return list(descriptors)

    @root_validator(pre=False, skip_on_failure=True)
    def validate_list_lengths(cls, values):
        """compares the length of the defined descriptors list with the provided values

        Args:
            values (Dict): Dictionary with all attribues

        Raises:
            ValueError: when the number of descriptors does not math the number of provided values

        Returns:
            Dict: Dict with the attributes
        """
        if len(values["descriptors"]) != len(values["values"]):
            raise ValueError(
                'must provide same number of descriptors and values, got {len(values["descriptors"])} != {len(values["values"])}'
            )
        return values

    def to_df(self) -> pd.DataFrame:
        """tabular overview of the feature as DataFrame

        Returns:
            pd.DataFrame: tabular overview of the feature as DataFrame
        """
        return pd.DataFrame(
            data=[self.values], index=[self.key], columns=self.descriptors
        )


class CategoricalDescriptorInput(CategoricalInput):
    """Class for categorical input features with descriptors

    Attributes:
        categories (List[str]): Names of the categories.
        allowed (List[bool]): List of bools indicating if a category is allowed within the optimization.
        descriptors (List[str]): List of strings representing the names of the descriptors.
        values (List[List[float]]): List of lists representing the descriptor values.
    """

    type: Literal["CategoricalDescriptorInput"] = "CategoricalDescriptorInput"
    order_id: ClassVar[int] = 4

    descriptors: TDescriptors
    values: TCategoricalDescriptorVals

    @validator("descriptors")
    def validate_descriptors(cls, descriptors):
        """validates that descriptors have unique names

        Args:
            categories (List[str]): List of descriptor names

        Raises:
            ValueError: when descriptors have non-unique names

        Returns:
            List[str]: List of the descriptors
        """
        descriptors = list(descriptors)
        if len(descriptors) != len(set(descriptors)):
            raise ValueError("descriptors must be unique")
        return descriptors

    @validator("values")
    def validate_values(cls, v, values):
        """validates the compatability of passed values for the descriptors and the defined categories

        Args:
            v (List[List[float]]): Nested list with descriptor values
            values (Dict): Dictionary with attributes

        Raises:
            ValueError: when values have different length than categories
            ValueError: when rows in values have different length than descriptors
            ValueError: when a descriptor shows no variance in the data

        Returns:
            List[List[float]]: Nested list with descriptor values
        """
        if len(v) != len(values["categories"]):
            raise ValueError("values must have same length as categories")
        for row in v:
            if len(row) != len(values["descriptors"]):
                raise ValueError("rows in values must have same length as descriptors")
        a = np.array(v)
        for i, d in enumerate(values["descriptors"]):
            if len(set(a[:, i])) == 1:
                raise ValueError(f"No variation for descriptor {d}.")
        return v

    @staticmethod
    def valid_transform_types() -> List[CategoricalEncodingEnum]:
        return [
            CategoricalEncodingEnum.ONE_HOT,
            CategoricalEncodingEnum.DUMMY,
            CategoricalEncodingEnum.ORDINAL,
            CategoricalEncodingEnum.DESCRIPTOR,
        ]

    def to_df(self):
        """tabular overview of the feature as DataFrame

        Returns:
            pd.DataFrame: tabular overview of the feature as DataFrame
        """
        data = dict(zip(self.categories, self.values))
        return pd.DataFrame.from_dict(data, orient="index", columns=self.descriptors)

    def fixed_value(
        self, transform_type: Optional[TTransform] = None
    ) -> Union[List[str], List[float], None]:
        """Returns the categories to which the feature is fixed, None if the feature is not fixed

        Returns:
            List[str]: List of categories or None
        """
        if transform_type != CategoricalEncodingEnum.DESCRIPTOR:
            return super().fixed_value(transform_type)
        else:
            val = self.get_allowed_categories()[0]
            return self.to_descriptor_encoding(pd.Series([val])).values[0].tolist()

    def get_bounds(
        self, transform_type: TTransform, values: Optional[pd.Series] = None
    ) -> Tuple[List[float], List[float]]:
        if transform_type != CategoricalEncodingEnum.DESCRIPTOR:
            return super().get_bounds(transform_type, values)
        else:
            # in case that values is None, we return the optimization bounds
            # else we return the complete bounds
            if values is None:
                df = self.to_df().loc[self.get_allowed_categories()]
            else:
                df = self.to_df()
            lower = df.min().values.tolist()  # type: ignore
            upper = df.max().values.tolist()  # type: ignore
            return lower, upper

    def validate_experimental(
        self, values: pd.Series, strict: bool = False
    ) -> pd.Series:
        """Method to validate the experimental dataFrame

        Args:
            values (pd.Series): A dataFrame with experiments
            strict (bool, optional): Boolean to distinguish if the occurence of fixed features in the dataset should be considered or not. Defaults to False.

        Raises:
            ValueError: when an entry is not in the list of allowed categories
            ValueError: when there is no variation in a feature provided by the experimental data
            ValueError: when no variation is present or planed for a given descriptor

        Returns:
            pd.Series: A dataFrame with experiments
        """
        values = super().validate_experimental(values, strict)
        if strict:
            lower, upper = self.get_bounds(
                transform_type=CategoricalEncodingEnum.DESCRIPTOR, values=values
            )
            for i, desc in enumerate(self.descriptors):
                if lower[i] == upper[i]:
                    raise ValueError(
                        f"No variation present or planned for descriptor {desc} for feature {self.key}. Remove the descriptor."
                    )
        return values

    @classmethod
    def from_df(cls, key: str, df: pd.DataFrame):
        """Creates a feature from a dataframe

        Args:
            key (str): The name of the feature
            df (pd.DataFrame): Categories as rows and descriptors as columns

        Returns:
            _type_: _description_
        """
        return cls(
            key=key,
            categories=list(df.index),
            allowed=[True for _ in range(len(df))],
            descriptors=list(df.columns),
            values=df.values.tolist(),
        )

    def to_descriptor_encoding(self, values: pd.Series) -> pd.DataFrame:
        """Converts values to descriptor encoding.

        Args:
            values (pd.Series): Values to transform.

        Returns:
            pd.DataFrame: Descriptor encoded dataframe.
        """
        return pd.DataFrame(
            data=values.map(
                dict(zip(self.categories, self.values))
            ).values.tolist(),  # type: ignore
            columns=[f"{self.key}{_CAT_SEP}{d}" for d in self.descriptors],
            index=values.index,
        )

    def from_descriptor_encoding(self, values: pd.DataFrame) -> pd.Series:
        """Converts values back from descriptor encoding.

        Args:
            values (pd.DataFrame): Descriptor encoded dataframe.

        Raises:
            ValueError: If descriptor columns not found in the dataframe.

        Returns:
            pd.Series: Series with categorical values.
        """
        cat_cols = [f"{self.key}{_CAT_SEP}{d}" for d in self.descriptors]
        # we allow here explicitly that the dataframe can have more columns than needed to have it
        # easier in the backtransform.
        if np.any([c not in values.columns for c in cat_cols]):
            raise ValueError(
                f"{self.key}: Column names don't match categorical levels: {values.columns}, {cat_cols}."
            )
        s = pd.DataFrame(
            data=np.sqrt(
                np.sum(
                    (
                        values[cat_cols].to_numpy()[:, np.newaxis, :]
                        - self.to_df().iloc[self.allowed].to_numpy()
                    )
                    ** 2,
                    axis=2,
                )
            ),
            columns=self.get_allowed_categories(),
            index=values.index,
        ).idxmin(1)
        s.name = self.key
        return s



import warnings
from typing import Literal, Optional

import numpy as np
import pandas as pd
from pydantic import validator

from bofire.data_models.constraints.constraint import Constraint, FeatureKeys


class NonlinearConstraint(Constraint):
    """Base class for nonlinear equality and inequality constraints.

    Attributes:
        expression (str): Mathematical expression that can be evaluated by `pandas.eval`.
        jacobian_expression (str): Mathematical expression that that can be evaluated by `pandas.eval`.
        features (list): list of feature keys (str) on which the constraint works on.
    """

    expression: str
    features: Optional[FeatureKeys] = None
    jacobian_expression: Optional[str] = None

    @validator("jacobian_expression", always=True)
    def set_jacobian_expression(cls, jacobian_expression, values):
        try:
            import sympy  # type: ignore
        except ImportError as e:
            warnings.warn(e.msg)
            warnings.warn("please run `pip install sympy` for this functionality.")
            return jacobian_expression

        if (
            jacobian_expression is None
            and "features" in values
            and "expression" in values
        ):
            if values["features"] is not None:
                return (
                    "["
                    + ", ".join(
                        [
                            str(sympy.S(values["expression"]).diff(key))
                            for key in values["features"]
                        ]
                    )
                    + "]"
                )
        return jacobian_expression

    def __call__(self, experiments: pd.DataFrame) -> pd.Series:
        return experiments.eval(self.expression)

    def jacobian(self, experiments: pd.DataFrame) -> pd.DataFrame:
        if self.jacobian_expression is not None:
            res = experiments.eval(self.jacobian_expression)
            for i, col in enumerate(res):
                if not hasattr(col, "__iter__"):
                    res[i] = pd.Series(np.repeat(col, experiments.shape[0]))

            if self.features is not None:
                return pd.DataFrame(
                    res, index=["dg/d" + name for name in self.features]
                ).transpose()
            else:
                return pd.DataFrame(
                    res, index=[f"dg/dx{i}" for i in range(experiments.shape[1])]
                ).transpose()

        raise ValueError(
            "The jacobian of a nonlinear constraint cannot be evaluated if jacobian_expression is None."
        )


class NonlinearEqualityConstraint(NonlinearConstraint):
    """Nonlinear equality constraint of the form 'expression == 0'.

    Attributes:
        expression: Mathematical expression that can be evaluated by `pandas.eval`.
    """

    type: Literal["NonlinearEqualityConstraint"] = "NonlinearEqualityConstraint"

    def is_fulfilled(self, experiments: pd.DataFrame, tol: float = 1e-6) -> pd.Series:
        return pd.Series(
            np.isclose(self(experiments), 0, atol=tol), index=experiments.index
        )

    def __str__(self):
        return f"{self.expression}==0"


class NonlinearInequalityConstraint(NonlinearConstraint):
    """Nonlinear inequality constraint of the form 'expression <= 0'.

    Attributes:
        expression: Mathematical expression that can be evaluated by `pandas.eval`.
    """

    type: Literal["NonlinearInequalityConstraint"] = "NonlinearInequalityConstraint"

    def is_fulfilled(self, experiments: pd.DataFrame, tol: float = 1e-6) -> pd.Series:
        return self(experiments) <= 0 + tol

    def __str__(self):
        return f"{self.expression}<=0"



from typing import Literal, Optional

from bofire.data_models.kernels.kernel import Kernel
from bofire.data_models.priors.api import AnyPrior


class ContinuousKernel(Kernel):
    pass


class RBFKernel(ContinuousKernel):
    type: Literal["RBFKernel"] = "RBFKernel"
    ard: bool = True
    lengthscale_prior: Optional[AnyPrior] = None


class MaternKernel(ContinuousKernel):
    type: Literal["MaternKernel"] = "MaternKernel"
    ard: bool = True
    nu: float = 2.5
    lengthscale_prior: Optional[AnyPrior] = None


class LinearKernel(ContinuousKernel):
    type: Literal["LinearKernel"] = "LinearKernel"
    variance_prior: Optional[AnyPrior] = None


class PolynomialKernel(ContinuousKernel):
    type: Literal["PolynomialKernel"] = "PolynomialKernel"
    offset_prior: Optional[AnyPrior] = None
    power: int = 2



from typing import List, Literal, Tuple

import numpy as np
import pandas as pd
from pydantic import root_validator, validator

from bofire.data_models.constraints.constraint import (
    Coefficients,
    Constraint,
    FeatureKeys,
)


class LinearConstraint(Constraint):
    """Abstract base class for linear equality and inequality constraints.

    Attributes:
        features (list): list of feature keys (str) on which the constraint works on.
        coefficients (list): list of coefficients (float) of the constraint.
        rhs (float): Right-hand side of the constraint
    """

    type: Literal["LinearConstraint"] = "LinearConstraint"

    features: FeatureKeys
    coefficients: Coefficients
    rhs: float

    @validator("features")
    def validate_features_unique(cls, features):
        """Validate that feature keys are unique."""
        if len(features) != len(set(features)):
            raise ValueError("features must be unique")
        return features

    @root_validator(pre=False, skip_on_failure=True)
    def validate_list_lengths(cls, values):
        """Validate that length of the feature and coefficient lists have the same length."""
        if len(values["features"]) != len(values["coefficients"]):
            raise ValueError(
                f'must provide same number of features and coefficients, got {len(values["features"])} != {len(values["coefficients"])}'
            )
        return values

    def __call__(self, experiments: pd.DataFrame) -> pd.Series:
        return (
            experiments[self.features] @ self.coefficients - self.rhs
        ) / np.linalg.norm(self.coefficients)

    # def lhs(self, df_data: pd.DataFrame) -> float:
    #     """Evaluate the left-hand side of the constraint on each row of a dataframe

    #     Args:
    #         df_data (pd.DataFrame): Dataframe on which the left-hand side should be evaluated.

    #     Returns:
    #         np.array: 1-dim array with left-hand side of each row of the provided dataframe.
    #     """
    #     cols = self.features
    #     coefficients = self.coefficients
    #     return np.sum(df_data[cols].values * np.array(coefficients), axis=1)

    def __str__(self) -> str:
        """Generate string representation of the constraint.

        Returns:
            str: string representation of the constraint.
        """
        return " + ".join(
            [f"{self.coefficients[i]} * {feat}" for i, feat in enumerate(self.features)]
        )

    def jacobian(self, experiments: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            np.tile(
                [np.array(self.coefficients) / np.linalg.norm(self.coefficients)],
                [experiments.shape[0], 1],
            ),
            columns=[f"dg/d{name}" for name in self.features],
        )


class LinearEqualityConstraint(LinearConstraint):
    """Linear equality constraint of the form `coefficients * x = rhs`.

    Attributes:
        features (list): list of feature keys (str) on which the constraint works on.
        coefficients (list): list of coefficients (float) of the constraint.
        rhs (float): Right-hand side of the constraint
    """

    type: Literal["LinearEqualityConstraint"] = "LinearEqualityConstraint"

    # def is_fulfilled(self, experiments: pd.DataFrame, complete: bool) -> bool:
    #     """Check if the linear equality constraint is fulfilled for all the rows of the provided dataframe.

    #     Args:
    #         df_data (pd.DataFrame): Dataframe to evaluate constraint on.

    #     Returns:
    #         bool: True if fulfilled else False.
    #     """
    #     fulfilled = np.isclose(self(experiments), 0)
    #     if complete:
    #         return fulfilled.all()
    #     else:
    #         pd.Series(fulfilled, index=experiments.index)

    def is_fulfilled(self, experiments: pd.DataFrame, tol: float = 1e-6) -> pd.Series:
        return pd.Series(
            np.isclose(self(experiments), 0, atol=tol), index=experiments.index
        )

    def __str__(self) -> str:
        """Generate string representation of the constraint.

        Returns:
            str: string representation of the constraint.
        """
        return super().__str__() + f" = {self.rhs}"


class LinearInequalityConstraint(LinearConstraint):
    """Linear inequality constraint of the form `coefficients * x <= rhs`.

    To instantiate a constraint of the form `coefficients * x >= rhs` multiply coefficients and rhs by -1, or
    use the classmethod `from_greater_equal`.

    Attributes:
        features (list): list of feature keys (str) on which the constraint works on.
        coefficients (list): list of coefficients (float) of the constraint.
        rhs (float): Right-hand side of the constraint
    """

    type: Literal["LinearInequalityConstraint"] = "LinearInequalityConstraint"

    def is_fulfilled(self, experiments: pd.DataFrame, tol: float = 1e-6) -> pd.Series:
        return self(experiments) <= 0 + tol

    def as_smaller_equal(self) -> Tuple[List[str], List[float], float]:
        """Return attributes in the smaller equal convention

        Returns:
            Tuple[List[str], List[float], float]: features, coefficients, rhs
        """
        return self.features, self.coefficients, self.rhs

    def as_greater_equal(self) -> Tuple[List[str], List[float], float]:
        """Return attributes in the greater equal convention

        Returns:
            Tuple[List[str], List[float], float]: features, coefficients, rhs
        """
        return self.features, [-1.0 * c for c in self.coefficients], -1.0 * self.rhs

    @classmethod
    def from_greater_equal(
        cls,
        features: List[str],
        coefficients: List[float],
        rhs: float,
    ):
        """Class method to construct linear inequality constraint of the form `coefficients * x >= rhs`.

        Args:
            features (List[str]): List of feature keys.
            coefficients (List[float]): List of coefficients.
            rhs (float): Right-hand side of the constraint.
        """
        return cls(
            features=features,
            coefficients=[-1.0 * c for c in coefficients],
            rhs=-1.0 * rhs,
        )

    @classmethod
    def from_smaller_equal(
        cls,
        features: List[str],
        coefficients: List[float],
        rhs: float,
    ):
        """Class method to construct linear inequality constraint of the form `coefficients * x <= rhs`.

        Args:
            features (List[str]): List of feature keys.
            coefficients (List[float]): List of coefficients.
            rhs (float): Right-hand side of the constraint.
        """
        return cls(
            features=features,
            coefficients=coefficients,
            rhs=rhs,
        )

    def __str__(self):
        """Generate string representation of the constraint.

        Returns:
            str: string representation of the constraint.
        """
        return super().__str__() + f" <= {self.rhs}"



from typing import Optional

from pydantic import Field, validator

from bofire.data_models.base import BaseModel
from bofire.data_models.domain.api import Inputs, Outputs
from bofire.data_models.features.api import TInputTransformSpecs


class Surrogate(BaseModel):
    type: str

    inputs: Inputs
    outputs: Outputs
    input_preprocessing_specs: TInputTransformSpecs = Field(default_factory=dict)
    dump: Optional[str] = None

    @validator("input_preprocessing_specs", always=True)
    def validate_input_preprocessing_specs(cls, v, values):
        # we also validate the number of input features here
        if len(values["inputs"]) == 0:
            raise ValueError("At least one input feature has to be provided.")
        v = values["inputs"]._validate_transform_specs(v)
        return v

    @validator("outputs")
    def validate_outputs(cls, v, values):
        if len(v) == 0:
            raise ValueError("At least one output feature has to be provided.")
        return v



from bofire.strategies import api as strategies


def test_strategy_can_be_loaded_from_data_model(strategy_spec):
    data_model = strategy_spec.obj()
    strategy = strategies.map(data_model=data_model)
    assert strategy is not None



import pandas as pd

from bofire.benchmarks.benchmark import Benchmark
from bofire.data_models.domain.api import Domain


class LookupTableBenchmark(Benchmark):
    """This class is a benchmark for discrete data where you can lookup a table for new inputs.
    It takes in a file with inputs and outputs and returns output for matching inputs from that file.

    Args:
        Benchmark: Subclass of the Benchmark function class.
    """

    def __init__(
        self,
        domain: Domain,
        lookup_table: pd.DataFrame,
        **kwargs,
    ) -> None:
        """Initializes Molecule_benchmark. A class that connects to categorical molecular input.

        Args:
            domain (Domain): Domain of the inputs and outputs
            LookUpTable (pd.DataFrame): DataFrame containing the LookUp table.
        """
        super().__init__(**kwargs)

        self._domain = domain
        self.lookup_table = lookup_table
        self.domain.validate_experiments(self.lookup_table)

    def _f(self, sampled: pd.DataFrame, **kwargs) -> pd.DataFrame:
        """return output values for matching SMILE candidates.

        Args:
            sampled (pd.DataFrame): Input values with input columns only.

        Returns:
            pd.DataFrame: output values from the LookUpTable. Columns are ouput keys and valid_output keys.
        """
        X = sampled.copy()
        X["proxy_index"] = X.index
        X_temp = pd.merge(
            X, self.lookup_table, on=self.domain.inputs.get_keys(), how="left"
        ).dropna()
        df = pd.merge(X, X_temp, how="left", indicator=True)
        if (
            df.loc[df._merge == "left_only", df.columns != "_merge"].proxy_index.size
            != 0
        ):
            indices = df.loc[
                df._merge == "left_only", df.columns != "_merge"
            ].proxy_index.to_list()
            raise ValueError(f"Input combination {indices} not found in Look up table")
        Y = X_temp[
            self.domain.outputs.get_keys()
            + [f"valid_{k}" for k in self.domain.outputs.get_keys()]
        ]
        Y.index = pd.RangeIndex(len(Y))
        return Y



from typing import Dict, Optional, Sequence

import numpy as np
import pandas as pd

from bofire.data_models.enum import RegressionMetricsEnum
from bofire.surrogates.diagnostics import metrics
from bofire.surrogates.single_task_gp import SingleTaskGPSurrogate
from bofire.surrogates.surrogate import Surrogate


def lengthscale_importance(surrogate: SingleTaskGPSurrogate) -> pd.Series:
    """Compute the lengthscale importance based on ARD.

    Args:
        surrogate (SingleTaskGPSurrogate): Surrogate to extract the importances.

    Returns:
        pd.Series: The importance values (inverse of the individual lenght scales).
    """
    try:
        scales = surrogate.model.covar_module.base_kernel.lengthscale  # type: ignore
    except AttributeError:
        raise ValueError("No lenghtscale based kernel found.")
    scales = 1.0 / scales.squeeze().detach().numpy()  # type: ignore
    if isinstance(scales, float):
        raise ValueError("Only one lengthscale found, use `ard=True`.")
    if len(scales) != len(surrogate.inputs):
        raise ValueError(
            "Number of lengthscale parameters to not matches the number of inputs."
        )
    return pd.Series(data=scales, index=surrogate.inputs.get_keys())


def lengthscale_importance_hook(
    surrogate: SingleTaskGPSurrogate,
    X_train: Optional[pd.DataFrame] = None,
    y_train: Optional[pd.DataFrame] = None,
    X_test: Optional[pd.DataFrame] = None,
    y_test: Optional[pd.DataFrame] = None,
):
    """Hook that can be used within `model.cross_validate` to compute a cross validated permutation feature importance."""
    return lengthscale_importance(surrogate=surrogate)


def combine_lengthscale_importances(importances: Sequence[pd.Series]) -> pd.DataFrame:
    """Combine the importance values from each fold into one dataframe.

    Args:
        importances (Sequence[pd.Series]): List of importance values per fold.

    Returns:
        pd.DataFrame: Dataframe with feature keys as columns, and one row per fold.
    """
    return pd.concat(importances, axis=1).T


def permutation_importance(
    surrogate: Surrogate,
    X: pd.DataFrame,
    y: pd.DataFrame,
    n_repeats: int = 5,
    seed: int = 42,
) -> Dict[str, pd.DataFrame]:
    """Computes permutation feature importance for a model.

    Args:
        model (Model): Model for which the feature importances should be estimated.
        X (pd.DataFrame): X values used to estimate the importances.
        y (pd.DataFrame): Y values used to estimate the importances.
        n_repeats (int, optional): Number of repeats. Defaults to 5.
        seed (int, optional): Seed for the random sampler. Defaults to 42.

    Returns:
        Dict[str, pd.DataFrame]: keys are the metrices for which the model is evluated and value is a dataframe
            with the feature keys as columns and the mean and std of the respective permutation importances as rows.
    """
    assert len(surrogate.outputs) == 1, "Only single output model supported so far."
    assert n_repeats > 1, "Number of repeats has to be larger than 1."
    assert seed > 0, "Seed has to be larger than zero."

    signs = {
        RegressionMetricsEnum.R2: 1.0,
        RegressionMetricsEnum.FISHER: -1.0,
        RegressionMetricsEnum.MAE: -1.0,
        RegressionMetricsEnum.MAPE: -1.0,
        RegressionMetricsEnum.MSD: -1.0,
        RegressionMetricsEnum.PEARSON: 1.0,
        RegressionMetricsEnum.SPEARMAN: 1.0,
    }

    output_key = surrogate.outputs[0].key
    rng = np.random.default_rng(seed)
    prelim_results = {
        k.name: {feature.key: [] for feature in surrogate.inputs}
        for k in metrics.keys()
    }
    pred = surrogate.predict(X)
    if len(pred) >= 2:
        original_metrics = {
            k.name: metrics[k](y[output_key].values, pred[output_key + "_pred"].values)  # type: ignore
            for k in metrics.keys()
        }
    else:
        original_metrics = {k.name: np.nan for k in metrics.keys()}

    for feature in surrogate.inputs:
        for _ in range(n_repeats):
            # shuffle
            X_i = X.copy()
            X_i[feature.key] = rng.permutation(X_i[feature.key].values)  # type: ignore
            # predict
            pred = surrogate.predict(X_i)
            # compute scores
            for metricenum, metric in metrics.items():
                if len(pred) >= 2:
                    prelim_results[metricenum.name][feature.key].append(
                        metric(y[output_key].values, pred[output_key + "_pred"].values)  # type: ignore
                    )
                else:
                    prelim_results[metricenum.name][feature.key].append(np.nan)  # type: ignore

    # convert dictionaries to dataframe for easier postprocessing and statistics
    # and return
    results = {}
    for k in metrics.keys():
        results[k.name] = pd.DataFrame(
            data={
                feature.key: [
                    original_metrics[k.name]
                    - np.mean(prelim_results[k.name][feature.key]),
                    np.std(prelim_results[k.name][feature.key]),
                ]
                for feature in surrogate.inputs
            },
            index=["mean", "std"],
        )
        results[k.name].loc["mean"] *= signs[k]

    return results


def permutation_importance_hook(
    surrogate: Surrogate,
    X_train: pd.DataFrame,
    y_train: pd.DataFrame,
    X_test: pd.DataFrame,
    y_test: pd.DataFrame,
    use_test: bool = True,
    n_repeats: int = 5,
    seed: int = 42,
) -> Dict[str, pd.DataFrame]:
    """Hook that can be used within `model.cross_validate` to compute a cross validated permutation feature importance.

    Args:
        model (Model): Predictive BoFire model.
        X_train (pd.DataFrame): Current train fold. X values.
        y_train (pd.DataFrame): Current train fold. y values.
        X_test (pd.DataFrame): Current test fold. X values.
        y_test (pd.DataFrame): Current test fold. y values.
        use_test (bool, optional): If True test fold is used to calculate feature importance else train fold is used.
            Defaults to True.
        n_repeats (int, optional): Number of repeats per feature. Defaults to 5.
        seed (int, optional): Seed for the random number generator. Defaults to 42.

    Returns:
        Dict[str, pd.DataFrame]: keys are the metrices for which the model is evluated and value is a dataframe
            with the feature keys as columns and the mean and std of the respective permutation importances as rows.
    """
    if use_test:
        X = X_test
        y = y_test
    else:
        X = X_train
        y = y_train
    return permutation_importance(
        surrogate=surrogate, X=X, y=y, n_repeats=n_repeats, seed=seed
    )


def combine_permutation_importances(
    importances: Sequence[Dict[str, pd.DataFrame]],
    metric: RegressionMetricsEnum = RegressionMetricsEnum.R2,
) -> pd.DataFrame:
    """Combines feature importances of a set of folds into one data frame for a requested metric.

    Args:
        importances (List[Dict[str, pd.DataFrame]]): List of permutation importance dictionaries, one per fold.
        metric (RegressionMetricsEnum, optional): Metric for which the data should be combined.
            Defaults to RegressionMetricsEnum.R2

    Returns:
        pd.DataFrame: Dataframe holding the mean permutation importance per fold and feature. Can be further processed by
            `describe`.
    """
    feature_keys = importances[0]["MAE"].columns
    return pd.DataFrame(
        data={
            key: [fold[metric.name].loc["mean", key] for fold in importances]
            for key in feature_keys
        }
    )



import os

import pytest

from bofire.benchmarks.aspen_benchmark import Aspen_benchmark
from bofire.data_models.domain.api import Domain, Inputs, Outputs
from bofire.data_models.features.api import (
    CategoricalDescriptorInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.objectives.api import MaximizeObjective


@pytest.mark.parametrize(
    "cls_benchmark",
    [
        (Aspen_benchmark),
    ],
)
def test_aspen_benchmark(cls_benchmark: Aspen_benchmark):
    """Tests the initializer of Aspen_benchmark and whether the filename, domain and paths are set up correctly.

    Args:
        cls_benchmark (Aspen_benchmark): Aspen_benchmark class
        return_complete (bool): _description_
        kwargs (dict): Arguments to the initializer of Aspen_benchmark. {"filename": , "domain": , "paths": }
    """

    domain = Domain(
        inputs=Inputs(
            features=[
                ContinuousInput(key="A", type="ContinuousInput", bounds=(0, 10)),
                ContinuousInput(
                    key="B",
                    type="ContinuousInput",
                    bounds=(-45, 20),
                ),
                CategoricalDescriptorInput(
                    key="C",
                    type="CategoricalDescriptorInput",
                    categories=["0", "1"],
                    allowed=[True, True],
                    descriptors=["binary"],
                    values=[[0.0], [1.0]],
                ),
            ]
        ),
        outputs=Outputs(
            features=[
                ContinuousOutput(
                    key="X",
                    type="ContinuousOutput",
                    objective=MaximizeObjective(
                        type="MaximizeObjective",
                        w=1.0,
                    ),
                )
            ]
        ),
    )
    filename = "aspen_benchmark_test_file.apwz"
    paths = {
        "A": "\\Data\\Flowsheeting Options\\Calculator\\CA-01\\Input\\FVN_INIT_VAL\\A",
        "B": "\\Data\\Flowsheeting Options\\Calculator\\CA-01\\Input\\FVN_INIT_VAL\\B",
        "C": "\\Data\\Streams\\Stream1\\Input\\TEMP\\C",
        "X": "\\Data\\Blocks\\DUM-03\\Input\\X",
    }
    # Create an aspen mockup file, so an object of aspen_benchmark can be created.
    if not os.path.exists(filename):
        file = open(filename, "w")
        file.close()

    # Test, if domain gets set up correctly
    benchmark_function = cls_benchmark(filename=filename, domain=domain, paths=paths)
    assert benchmark_function.domain == domain, (
        "Error during domain set up."
        + "\n Expected: "
        + str(domain)
        + "\n Got: "
        + str(benchmark_function.domain)
    )
    assert benchmark_function.paths == paths, (
        "Error during set up of paths."
        + "\n Expected: "
        + str(paths)
        + "\n Got: "
        + str(benchmark_function.paths)
    )

    # Test, whether ValueError is raised, if a key has no path to aspen provided.
    paths.popitem()
    with pytest.raises(ValueError):
        benchmark_function = cls_benchmark(
            filename=filename, domain=domain, paths=paths
        )

    os.remove(filename)

    # Test, if filename error gets thrown for wrong or non-existent filename.
    with pytest.raises(ValueError):
        benchmark_function = cls_benchmark(
            filename=filename, domain=domain, paths=paths
        )


test_aspen_benchmark(cls_benchmark=Aspen_benchmark)



import numpy as np
import pandas as pd
import pytest
from scipy.stats import pearsonr, spearmanr
from sklearn.metrics import (
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    r2_score,
)

from bofire.data_models.enum import RegressionMetricsEnum
from bofire.data_models.features.api import CategoricalInput, ContinuousInput
from bofire.surrogates.diagnostics import (
    CvResult,
    CvResults,
    CvResults2CrossValidationValues,
    UQ_metrics,
    _CVPPDiagram,
    _mean_absolute_error,
    _mean_absolute_percentage_error,
    _mean_squared_error,
    _pearson,
    _r2_score,
    _spearman,
    metrics,
)


def generate_cvresult(
    key,
    n_samples,
    include_labcodes=False,
    include_X=False,
    include_standard_deviation=False,
):
    feature = ContinuousInput(
        key=key,
        bounds=(10, 20),
    )
    observed = feature.sample(n_samples)
    predicted = observed + np.random.normal(0, 1, size=n_samples)
    if include_labcodes:
        labcodes = pd.Series([str(i) for i in range(n_samples)])
    else:
        labcodes = None
    if include_X:
        X = pd.DataFrame(
            data=np.random.uniform(size=(n_samples, 2)), columns=["a", "b"]
        )
    else:
        X = None
    if include_standard_deviation:
        standard_deviation = pd.Series(np.random.normal(0, 0.2, size=n_samples))
    else:
        standard_deviation = None
    return CvResult(
        key=key,
        observed=observed,
        predicted=predicted,
        labcodes=labcodes,
        X=X,
        standard_deviation=standard_deviation,
    )


@pytest.mark.parametrize(
    "bofire, sklearn",
    [
        (_mean_absolute_error, mean_absolute_error),
        (_mean_absolute_percentage_error, mean_absolute_percentage_error),
        (_mean_squared_error, mean_squared_error),
        (_r2_score, r2_score),
    ],
)
def test_sklearn_metrics(bofire, sklearn):
    n_samples = 20
    feature = ContinuousInput(
        key="a",
        bounds=(10, 20),
    )
    observed = feature.sample(n_samples).values
    predicted = observed + np.random.normal(0, 1, size=n_samples)
    sd = np.random.normal(0, 1, size=n_samples)
    assert bofire(observed, predicted, sd) == sklearn(observed, predicted)
    assert bofire(observed, predicted) == sklearn(observed, predicted)


@pytest.mark.parametrize(
    "bofire, scipy",
    [
        (_pearson, pearsonr),
        (_spearman, spearmanr),
    ],
)
def test_scipy_metrics(bofire, scipy):
    n_samples = 20
    feature = ContinuousInput(
        key="a",
        bounds=(10, 20),
    )
    observed = feature.sample(n_samples).values
    predicted = observed + np.random.normal(0, 1, size=n_samples)
    sd = np.random.normal(0, 1, size=n_samples)
    s, _ = scipy(predicted, observed)
    assert bofire(observed, predicted, sd) == s
    assert bofire(observed, predicted) == s


def test_cvresult_not_numeric():
    n_samples = 10
    feature = ContinuousInput(
        key="a",
        bounds=(10, 20),
    )
    feature2 = CategoricalInput(key="a", categories=["a", "b"])
    with pytest.raises(ValueError):
        CvResult(
            key=feature.key,
            observed=feature2.sample(n_samples),
            predicted=feature.sample(n_samples),
        )
    with pytest.raises(ValueError):
        CvResult(
            key=feature.key,
            observed=feature.sample(n_samples),
            predicted=feature2.sample(n_samples),
        )
    with pytest.raises(ValueError):
        CvResult(
            key=feature.key,
            observed=feature.sample(n_samples),
            predicted=feature.sample(n_samples),
            standard_deviation=feature2.sample(n_samples),
        )


def test_cvresult_shape_mismatch():
    feature = ContinuousInput(
        key="a",
        bounds=(10, 20),
    )
    with pytest.raises(ValueError):
        CvResult(
            key=feature.key,
            observed=feature.sample(5),
            predicted=feature.sample(6),
        )
    with pytest.raises(ValueError):
        CvResult(
            key=feature.key,
            observed=feature.sample(5),
            predicted=feature.sample(5),
            standard_deviation=feature.sample(6),
        )
    with pytest.raises(ValueError):
        CvResult(
            key=feature.key,
            observed=feature.sample(5),
            predicted=feature.sample(5),
            standard_deviation=feature.sample(5),
            labcodes=pd.Series(["5", "6"]),
        )
    with pytest.raises(ValueError):
        CvResult(
            key=feature.key,
            observed=feature.sample(5),
            predicted=feature.sample(5),
            standard_deviation=feature.sample(5),
            labcodes=pd.Series(["5", "6", "7", "8", "9"]),
            X=pd.DataFrame(data=np.random.uniform(size=(2, 2)), columns=["a", "b"]),
        )
    CvResult(
        key=feature.key,
        observed=feature.sample(5),
        predicted=feature.sample(5),
        standard_deviation=feature.sample(5),
        labcodes=pd.Series(["5", "6", "7", "8", "9"]),
        X=pd.DataFrame(data=np.random.uniform(size=(5, 2)), columns=["a", "b"]),
    )


def test_cvresult_get_metric():
    n_samples = 10
    feature = ContinuousInput(
        key="a",
        bounds=(10, 20),
    )
    observed = feature.sample(n=n_samples)
    predicted = observed + np.random.normal(loc=0, scale=1, size=n_samples)
    cv = CvResult(key=feature.key, observed=observed, predicted=predicted)
    assert cv.n_samples == 10
    for metric in metrics.keys():
        cv.get_metric(metric)


def test_cvresult_get_metric_invalid():
    n_samples = 1
    feature = ContinuousInput(
        key="a",
        bounds=(10, 20),
    )
    observed = feature.sample(n=n_samples)
    predicted = observed + np.random.normal(loc=0, scale=1, size=n_samples)
    cv = CvResult(key=feature.key, observed=observed, predicted=predicted)
    for metric in metrics.keys():
        np.isnan(cv.get_metric(metric=metric))


def test_cvresult_get_UQ_metric_valid():
    cv = generate_cvresult(key="a", n_samples=10, include_standard_deviation=True)
    assert cv.n_samples == 10
    for metric in UQ_metrics.keys():
        m = cv.get_metric(metric=metric)
        assert type(m) == float


def test_cvresult_get_UQ_metric_invalid():
    cv = generate_cvresult(key="a", n_samples=10, include_standard_deviation=False)
    assert cv.n_samples == 10
    for metric in UQ_metrics.keys():
        with pytest.warns(UserWarning):
            np.isnan(cv.get_metric(metric=metric))


def test_CVPPDiagram_valid():
    cv = generate_cvresult(key="a", n_samples=10, include_standard_deviation=True)
    q, Cq = _CVPPDiagram(
        observed=cv.observed.values,
        predicted=cv.predicted.values,
        standard_deviation=cv.standard_deviation,
    )
    assert len(q) == len(Cq)


def test_CVPPDiagram_invalid():
    cv = generate_cvresult(key="a", n_samples=10, include_standard_deviation=False)
    with pytest.raises(
        ValueError,
        match="Calibration metric without standard deviation is not possible",
    ):
        _CVPPDiagram(
            observed=cv.observed.values,
            predicted=cv.predicted.values,
            standard_deviation=cv.standard_deviation,
        )


def test_cvresults_invalid():
    # test for empty results
    with pytest.raises(ValueError):
        CvResults(results=[])
    # test for wrong keys
    n_samples = 10
    feature = ContinuousInput(
        key="a",
        bounds=(10, 20),
    )
    observed = feature.sample(n=n_samples)
    predicted = observed + np.random.normal(loc=0, scale=1, size=n_samples)
    cv1 = CvResult(key=feature.key, observed=observed, predicted=predicted)
    cv2 = CvResult(key="b", observed=observed, predicted=predicted)
    with pytest.raises(ValueError):
        CvResults(results=[cv1, cv2])
    # test for missing standard_deviation
    cv1 = CvResult(key=feature.key, observed=observed, predicted=predicted)
    cv2 = CvResult(
        key=feature.key,
        observed=observed,
        predicted=predicted,
        standard_deviation=observed,
    )
    with pytest.raises(ValueError):
        CvResults(results=[cv1, cv2])
    # test for missing labcodes
    cv1 = CvResult(key=feature.key, observed=observed, predicted=predicted)
    cv2 = CvResult(
        key=feature.key,
        observed=observed,
        predicted=predicted,
        labcodes=pd.Series([str(i) for i in range(n_samples)]),
    )
    with pytest.raises(ValueError):
        CvResults(results=[cv1, cv2])
    # test for missing X
    cv1 = CvResult(key=feature.key, observed=observed, predicted=predicted)
    cv2 = CvResult(
        key=feature.key,
        observed=observed,
        predicted=predicted,
        X=pd.DataFrame(data=np.random.uniform(size=(10, 2)), columns=["a", "b"]),
    )
    with pytest.raises(ValueError):
        CvResults(results=[cv1, cv2])
    # test for wrong shape of X
    cv1 = CvResult(
        key=feature.key,
        observed=observed,
        predicted=predicted,
        X=pd.DataFrame(data=np.random.uniform(size=(10, 2)), columns=["a", "b"]),
    )
    cv2 = CvResult(
        key=feature.key,
        observed=observed,
        predicted=predicted,
        X=pd.DataFrame(data=np.random.uniform(size=(10, 3)), columns=["a", "b", "c"]),
    )
    with pytest.raises(ValueError):
        CvResults(results=[cv1, cv2])
    # working cvresult
    cv1 = CvResult(
        key=feature.key,
        observed=observed,
        predicted=predicted,
        labcodes=pd.Series([str(i) for i in range(n_samples)]),
        X=pd.DataFrame(data=np.random.uniform(size=(10, 2)), columns=["a", "b"]),
    )
    cv2 = CvResult(
        key=feature.key,
        observed=observed,
        predicted=predicted,
        labcodes=pd.Series([str(i) for i in range(n_samples)]),
        X=pd.DataFrame(data=np.random.uniform(size=(10, 2)), columns=["a", "b"]),
    )


@pytest.mark.parametrize(
    "cv_results",
    [
        CvResults(
            results=[generate_cvresult(key="a", n_samples=10) for _ in range(10)]
        ),
        CvResults(results=[generate_cvresult(key="a", n_samples=10) for _ in range(5)]),
    ],
)
def test_cvresults_get_metrics(cv_results):
    assert cv_results.key == "a"
    for metric in metrics:
        for combine_folds in [True, False]:
            m = cv_results.get_metric(metric, combine_folds)
            if combine_folds:
                assert len(m) == 1
            else:
                assert len(m) == len(cv_results.results)
            assert m.name == metric.name
    for combine_folds in [True, False]:
        df = cv_results.get_metrics(combine_folds=combine_folds)
        if combine_folds:
            assert df.shape == (1, len(metrics))
        else:
            assert df.shape == (len(cv_results.results), len(metrics))


def test_cvresults_get_metric_combine_folds():
    cv_results = CvResults(
        results=[generate_cvresult(key="a", n_samples=10) for _ in range(10)]
    )
    assert np.allclose(
        cv_results.get_metric(RegressionMetricsEnum.MAE, combine_folds=True).values[0],
        cv_results.get_metric(RegressionMetricsEnum.MAE, combine_folds=False).mean(),
    )


def test_cvresults_combine_folds():
    cv_results = CvResults(
        results=[
            generate_cvresult(
                key="a", n_samples=5, include_labcodes=True, include_X=True
            ),
            generate_cvresult(
                key="a", n_samples=6, include_labcodes=True, include_X=True
            ),
        ]
    )
    cv = cv_results._combine_folds()
    assert cv.observed.shape == (11,)
    assert cv.predicted.shape == (11,)
    assert cv.labcodes.shape == (11,)
    assert cv.X.shape == (11, 2)


@pytest.mark.parametrize(
    "cv_results",
    [
        CvResults(results=[generate_cvresult(key="a", n_samples=1) for _ in range(10)]),
        CvResults(results=[generate_cvresult(key="a", n_samples=1) for _ in range(5)]),
    ],
)
def test_cvresults_get_metrics_loo(cv_results):
    loocv = cv_results._combine_folds()
    assert loocv.observed.shape == (len(cv_results),)
    assert loocv.predicted.shape == (len(cv_results),)
    for metric in metrics:
        m = cv_results.get_metric(metric)
        assert len(m) == 1
    df = cv_results.get_metrics()
    assert df.shape == (1, len(metrics))


@pytest.mark.parametrize(
    "cv_results, expected",
    [
        (
            CvResults(
                results=[generate_cvresult(key="a", n_samples=1) for _ in range(5)]
            ),
            True,
        ),
        (
            CvResults(
                results=[generate_cvresult(key="a", n_samples=5) for _ in range(5)]
            ),
            False,
        ),
        (
            CvResults(
                results=[
                    generate_cvresult(key="a", n_samples=5),
                    generate_cvresult(key="a", n_samples=1),
                ]
            ),
            False,
        ),
    ],
)
def test_cvresults_is_loo(cv_results, expected):
    assert cv_results.is_loo == expected


@pytest.mark.parametrize(
    "cv_results",
    [
        CvResults(results=[generate_cvresult(key="a", n_samples=4) for _ in range(10)]),
        CvResults(
            results=[
                generate_cvresult(key="a", n_samples=6, include_standard_deviation=True)
                for _ in range(4)
            ]
        ),
    ],
)
def test_CvResults2CrossValidationValues(cv_results):
    metrics = cv_results.get_metrics(combine_folds=False)
    transformed = CvResults2CrossValidationValues(cv_results)
    assert isinstance(transformed, dict)
    assert list(transformed.keys()) == ["a"]
    assert len(transformed["a"]) == len(cv_results)
    for i in range(len(cv_results)):
        assert np.allclose(
            cv_results.results[i].predicted.values, transformed["a"][i].predicted
        )
        assert np.allclose(
            cv_results.results[i].observed.values, transformed["a"][i].observed
        )
        if cv_results.results[i].standard_deviation is not None:
            assert np.allclose(
                cv_results.results[i].standard_deviation.values,
                transformed["a"][i].standardDeviation,
            )
        else:
            assert transformed["a"][i].standardDeviation is None
        for m in metrics.columns:
            assert metrics.loc[i, m] == transformed["a"][i].metrics[m]


def test_CvResults2CrossValidationValues_minimal():
    cv_results = CvResults(
        results=[generate_cvresult(key="a", n_samples=2) for _ in range(4)]
        + [generate_cvresult(key="a", n_samples=1)]
    )
    transformed = CvResults2CrossValidationValues(cv_results)
    for i in range(5):
        if i < 4:
            assert transformed["a"][i].metrics is not None
        else:
            assert transformed["a"][i].metrics is None



from copy import deepcopy

import pytest

import bofire.strategies.api as strategies
from bofire.benchmarks.single import Himmelblau
from bofire.data_models.acquisition_functions.api import qNEI
from bofire.data_models.strategies.api import (
    AlwaysTrueCondition,
    NumberOfExperimentsCondition,
    RandomStrategy,
    SoboStrategy,
    Step,
    StepwiseStrategy,
)


def test_StepwiseStrategy_invalid_domains():
    benchmark = Himmelblau()
    domain2 = deepcopy(benchmark.domain)
    domain2.inputs[0].key = "mama"
    with pytest.raises(
        ValueError,
        match="Domain of step 0 is incompatible to domain of StepwiseStrategy.",
        # match=f"Domain of step {0} is incompatible to domain of StepwiseStrategy.",
    ):
        StepwiseStrategy(
            domain=benchmark.domain,
            steps=[
                Step(
                    strategy_data=RandomStrategy(domain=domain2),
                    condition=NumberOfExperimentsCondition(n_experiments=5),
                    max_parallelism=-1,
                ),
                Step(
                    strategy_data=SoboStrategy(
                        domain=benchmark.domain, acquisition_function=qNEI()
                    ),
                    condition=NumberOfExperimentsCondition(n_experiments=15),
                    max_parallelism=2,
                ),
            ],
        )


def test_StepwiseStrategy_invalid_AlwaysTrue():
    benchmark = Himmelblau()
    with pytest.raises(
        ValueError, match="`AlwaysTrueCondition` is only allowed for the last step."
    ):
        StepwiseStrategy(
            domain=benchmark.domain,
            steps=[
                Step(
                    strategy_data=RandomStrategy(domain=benchmark.domain),
                    condition=AlwaysTrueCondition(),
                    max_parallelism=-1,
                ),
                Step(
                    strategy_data=SoboStrategy(
                        domain=benchmark.domain, acquisition_function=qNEI()
                    ),
                    condition=NumberOfExperimentsCondition(n_experiments=10),
                    max_parallelism=2,
                ),
            ],
        )


@pytest.mark.parametrize(
    "n_experiments, expected_strategy, expected_index",
    [(5, RandomStrategy, 0), (10, SoboStrategy, 1)],
)
def test_StepWiseStrategy_get_step(n_experiments, expected_strategy, expected_index):
    benchmark = Himmelblau()
    experiments = benchmark.f(
        benchmark.domain.inputs.sample(n_experiments), return_complete=True
    )
    data_model = StepwiseStrategy(
        domain=benchmark.domain,
        steps=[
            Step(
                strategy_data=RandomStrategy(domain=benchmark.domain),
                condition=NumberOfExperimentsCondition(n_experiments=6),
                max_parallelism=-1,
            ),
            Step(
                strategy_data=SoboStrategy(
                    domain=benchmark.domain, acquisition_function=qNEI()
                ),
                condition=NumberOfExperimentsCondition(n_experiments=10),
                max_parallelism=2,
            ),
        ],
    )
    strategy = strategies.map(data_model)
    strategy.tell(experiments)
    i, step = strategy._get_step()
    assert isinstance(step.strategy_data, expected_strategy)
    assert i == expected_index


def test_StepWiseStrategy_get_step_invalid():
    benchmark = Himmelblau()
    experiments = benchmark.f(benchmark.domain.inputs.sample(12), return_complete=True)
    data_model = StepwiseStrategy(
        domain=benchmark.domain,
        steps=[
            Step(
                strategy_data=RandomStrategy(domain=benchmark.domain),
                condition=NumberOfExperimentsCondition(n_experiments=6),
                max_parallelism=-1,
            ),
            Step(
                strategy_data=SoboStrategy(
                    domain=benchmark.domain, acquisition_function=qNEI()
                ),
                condition=NumberOfExperimentsCondition(n_experiments=10),
                max_parallelism=2,
            ),
        ],
    )
    strategy = strategies.map(data_model)
    strategy.tell(experiments)
    with pytest.raises(ValueError, match="No condition could be satisfied."):
        strategy._get_step()


def test_StepWiseStrategy_invalid_ask():
    benchmark = Himmelblau()
    data_model = StepwiseStrategy(
        domain=benchmark.domain,
        steps=[
            Step(
                strategy_data=RandomStrategy(domain=benchmark.domain),
                condition=NumberOfExperimentsCondition(n_experiments=8),
                max_parallelism=2,
            ),
            Step(
                strategy_data=SoboStrategy(
                    domain=benchmark.domain, acquisition_function=qNEI()
                ),
                condition=NumberOfExperimentsCondition(n_experiments=10),
                max_parallelism=2,
            ),
        ],
    )
    strategy = strategies.map(data_model)
    experiments = benchmark.f(benchmark.domain.inputs.sample(2), return_complete=True)
    strategy.tell(experiments=experiments)
    with pytest.raises(
        ValueError, match="Maximum number of candidates for step 0 is 2."
    ):
        strategy.ask(3)


def test_StepWiseStrategy_ask():
    benchmark = Himmelblau()
    experiments = benchmark.f(benchmark.domain.inputs.sample(2), return_complete=True)
    data_model = StepwiseStrategy(
        domain=benchmark.domain,
        steps=[
            Step(
                strategy_data=RandomStrategy(domain=benchmark.domain),
                condition=NumberOfExperimentsCondition(n_experiments=5),
                max_parallelism=2,
            ),
            Step(
                strategy_data=SoboStrategy(
                    domain=benchmark.domain, acquisition_function=qNEI()
                ),
                condition=NumberOfExperimentsCondition(n_experiments=10),
                max_parallelism=2,
            ),
        ],
    )
    strategy = strategies.map(data_model)
    strategy.tell(experiments=experiments)
    candidates = strategy.ask(2)
    assert len(candidates) == 2



import sys
from itertools import combinations
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from formulaic import Formula
from scipy.optimize import LinearConstraint, NonlinearConstraint

from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
    NonlinearEqualityConstraint,
)
from bofire.data_models.constraints.nonlinear import NonlinearInequalityConstraint
from bofire.data_models.domain.domain import Domain
from bofire.data_models.features.continuous import ContinuousInput
from bofire.data_models.strategies.api import (
    PolytopeSampler as PolytopeSamplerDataModel,
)
from bofire.strategies.samplers.polytope import PolytopeSampler


def get_formula_from_string(
    model_type: Union[str, Formula] = "linear",
    domain: Optional[Domain] = None,
    rhs_only: bool = True,
) -> Formula:
    """Reformulates a string describing a model or certain keywords as Formula objects.

    Args:
        model_type (str or Formula): A formula containing all model terms.
        domain (Domain): A domain that nests necessary information on
        how to translate a problem to a formula. Contains a problem.
        rhs_only (bool): The function returns only the right hand side of the formula if set to True.
        Returns:
    A Formula object describing the model that was given as string or keyword.
    """
    # set maximum recursion depth to higher value
    recursion_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(2000)

    if isinstance(model_type, Formula):
        return model_type
        # build model if a keyword and a problem are given.
    else:
        # linear model
        if model_type == "linear":
            formula = linear_formula(domain)

        # linear and interactions model
        elif model_type == "linear-and-quadratic":
            formula = linear_and_quadratic_formula(domain)

        # linear and quadratic model
        elif model_type == "linear-and-interactions":
            formula = linear_and_interactions_formula(domain)

        # fully quadratic model
        elif model_type == "fully-quadratic":
            formula = fully_quadratic_formula(domain)

        else:
            formula = model_type + "   "

    formula = Formula(formula[:-3])

    if rhs_only:
        if hasattr(formula, "rhs"):
            formula = formula.rhs

    # set recursion limit to old value
    sys.setrecursionlimit(recursion_limit)

    return formula


def linear_formula(
    domain: Optional[Domain],
) -> str:
    """Reformulates a string describing a linear-model or certain keywords as Formula objects.
        formula = model_type + "   "

    Args: domain (Domain): A problem context that nests necessary information on
        how to translate a problem to a formula. Contains a problem.

    Returns:
        A string describing the model that was given as string or keyword.
    """
    assert (
        domain is not None
    ), "If the model is described by a keyword a domain must be provided"
    formula = "".join([input.key + " + " for input in domain.inputs])
    return formula


def linear_and_quadratic_formula(
    domain: Optional[Domain],
) -> str:
    """Reformulates a string describing a linear-and-quadratic model or certain keywords as Formula objects.

    Args: domain (Domain): A problem context that nests necessary information on
        how to translate a problem to a formula. Contains a problem.

    Returns:
        A string describing the model that was given as string or keyword.
    """
    assert (
        domain is not None
    ), "If the model is described by a keyword a problem must be provided."
    formula = "".join([input.key + " + " for input in domain.inputs])
    formula += "".join(["{" + input.key + "**2} + " for input in domain.inputs])
    return formula


def linear_and_interactions_formula(
    domain: Optional[Domain],
) -> str:
    """Reformulates a string describing a linear-and-interactions model or certain keywords as Formula objects.

    Args: domain (Domain): A problem context that nests necessary information on
        how to translate a problem to a formula. Contains a problem.

    Returns:
        A string describing the model that was given as string or keyword.
    """
    assert (
        domain is not None
    ), "If the model is described by a keyword a problem must be provided."
    formula = "".join([input.key + " + " for input in domain.inputs])
    for i in range(len(domain.inputs)):
        for j in range(i):
            formula += (
                domain.inputs.get_keys()[j] + ":" + domain.inputs.get_keys()[i] + " + "
            )
    return formula


def fully_quadratic_formula(
    domain: Optional[Domain],
) -> str:
    """Reformulates a string describing a fully-quadratic model or certain keywords as Formula objects.

    Args: domain (Domain): A problem context that nests necessary information on
        how to translate a problem to a formula. Contains a problem.

    Returns:
        A string describing the model that was given as string or keyword.
    """
    assert (
        domain is not None
    ), "If the model is described by a keyword a problem must be provided."
    formula = "".join([input.key + " + " for input in domain.inputs])
    for i in range(len(domain.inputs)):
        for j in range(i):
            formula += (
                domain.inputs.get_keys()[j] + ":" + domain.inputs.get_keys()[i] + " + "
            )
    formula += "".join(["{" + input.key + "**2} + " for input in domain.inputs])
    return formula


def n_zero_eigvals(
    domain: Domain, model_type: Union[str, Formula], epsilon=1e-7
) -> int:
    """Determine the number of eigenvalues of the information matrix that are necessarily zero because of
    equality constraints."""

    # sample points (fulfilling the constraints)
    model_formula = get_formula_from_string(
        model_type=model_type, rhs_only=True, domain=domain
    )
    N = len(model_formula) + 3

    sampler = PolytopeSampler(data_model=PolytopeSamplerDataModel(domain=domain))
    X = sampler.ask(N)
    # compute eigenvalues of information matrix
    A = model_formula.get_model_matrix(X)
    eigvals = np.abs(np.linalg.eigvalsh(A.T @ A))  # type: ignore

    return len(eigvals) - len(eigvals[eigvals > epsilon])


def constraints_as_scipy_constraints(
    domain: Domain,
    n_experiments: int,
    ignore_nchoosek: bool = True,
) -> List:
    """Formulates opti constraints as scipy constraints.

    Args:
        domain (Domain): Domain whose constraints should be formulated as scipy constraints.
        n_experiments (int): Number of instances of inputs for problem that are evaluated together.
        ingore_nchoosek (bool): NChooseK constraints are ignored if set to true. Defaults to True.

    Returns:
        A list of scipy constraints corresponding to the constraints of the given opti problem.
    """
    D = len(domain.inputs)

    # reformulate constraints
    constraints = []
    if len(domain.constraints) == 0:
        return constraints
    for c in domain.constraints:
        if isinstance(c, LinearEqualityConstraint):
            # write lower/upper bound as vector
            lb = np.ones(n_experiments) * (c.rhs / np.linalg.norm(c.coefficients))
            ub = np.ones(n_experiments) * (c.rhs / np.linalg.norm(c.coefficients))

            # write constraint as matrix
            lhs = {
                c.features[i]: c.coefficients[i] / np.linalg.norm(c.coefficients)
                for i in range(len(c.features))
            }
            row = np.zeros(D)
            for i, name in enumerate(domain.inputs.get_keys()):
                if name in lhs.keys():
                    row[i] = lhs[name]

            A = np.zeros(shape=(n_experiments, D * n_experiments))
            for i in range(n_experiments):
                A[i, i * D : (i + 1) * D] = row

            constraints.append(LinearConstraint(A, lb, ub))  # type: ignore

        elif isinstance(c, LinearInequalityConstraint):
            # write upper/lowe bound as vector
            lb = -np.inf * np.ones(n_experiments)
            ub = np.ones(n_experiments) * c.rhs / np.linalg.norm(c.coefficients)

            # write constraint as matrix
            lhs = {
                c.features[i]: c.coefficients[i] / np.linalg.norm(c.coefficients)
                for i in range(len(c.features))
            }
            row = np.zeros(D)
            for i, name in enumerate(domain.inputs.get_keys()):
                if name in lhs.keys():
                    row[i] = lhs[name]

            A = np.zeros(shape=(n_experiments, D * n_experiments))
            for i in range(n_experiments):
                A[i, i * D : (i + 1) * D] = row

            constraints.append(LinearConstraint(A, lb, ub))  # type: ignore

        elif isinstance(c, NonlinearEqualityConstraint):
            # write upper/lower bound as vector
            lb = np.zeros(n_experiments)
            ub = np.zeros(n_experiments)

            # define constraint evaluation (and gradient if provided)
            fun = ConstraintWrapper(
                constraint=c, domain=domain, n_experiments=n_experiments
            )

            if c.jacobian_expression is not None:
                constraints.append(NonlinearConstraint(fun, lb, ub, jac=fun.jacobian))
            else:
                constraints.append(NonlinearConstraint(fun, lb, ub))

        elif isinstance(c, NonlinearInequalityConstraint):
            # write upper/lower bound as vector
            lb = -np.inf * np.ones(n_experiments)
            ub = np.zeros(n_experiments)

            # define constraint evaluation (and gradient if provided)
            fun = ConstraintWrapper(
                constraint=c, domain=domain, n_experiments=n_experiments
            )

            if c.jacobian_expression is not None:
                constraints.append(NonlinearConstraint(fun, lb, ub, jac=fun.jacobian))
            else:
                constraints.append(NonlinearConstraint(fun, lb, ub))

        elif isinstance(c, NChooseKConstraint):
            if ignore_nchoosek:
                pass
            else:
                # write upper/lower bound as vector
                lb = -np.inf * np.ones(n_experiments)
                ub = np.zeros(n_experiments)

                # define constraint evaluation (and gradient if provided)
                fun = ConstraintWrapper(
                    constraint=c, domain=domain, n_experiments=n_experiments
                )

                constraints.append(NonlinearConstraint(fun, lb, ub, jac=fun.jacobian))

        else:
            raise NotImplementedError(f"No implementation for this constraint: {c}")

    return constraints


class ConstraintWrapper:
    """Wrapper for nonlinear constraints."""

    def __init__(
        self, constraint: NonlinearConstraint, domain: Domain, n_experiments: int = 0
    ) -> None:
        """
        Args:
            constraint (Constraint): constraint to be called
            domain (Domain): Domain the constraint belongs to
        """
        self.constraint = constraint
        self.names = domain.inputs.get_keys()
        self.D = len(domain.inputs)
        self.n_experiments = n_experiments
        if constraint.features is None:
            raise ValueError(
                f"The features attribute of constraint {constraint} is not set, but has to be set."
            )
        self.constraint_feature_indices = np.searchsorted(
            self.names, self.constraint.features
        )

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """call constraint with flattened numpy array."""
        x = pd.DataFrame(x.reshape(len(x) // self.D, self.D), columns=self.names)  # type: ignore
        violation = self.constraint(x).to_numpy()
        violation[np.abs(violation) < 0] = 0
        return violation  # type: ignore

    def jacobian(self, x: np.ndarray) -> np.ndarray:
        """call constraint gradient with flattened numpy array."""
        x = pd.DataFrame(x.reshape(len(x) // self.D, self.D), columns=self.names)
        gradient_compressed = self.constraint.jacobian(x).to_numpy()

        jacobian = np.zeros(shape=(self.n_experiments, self.D * self.n_experiments))
        rows = np.repeat(
            np.arange(self.n_experiments), len(self.constraint_feature_indices)
        )
        cols = np.repeat(
            self.D * np.arange(self.n_experiments), len(self.constraint_feature_indices)
        ).reshape((self.n_experiments, len(self.constraint_feature_indices)))
        cols = (cols + self.constraint_feature_indices).flatten()

        jacobian[rows, cols] = gradient_compressed.flatten()

        return jacobian


def d_optimality(X: np.ndarray, delta=1e-9) -> float:
    """Compute ln(1/|X^T X|) for a model matrix X (smaller is better).
    The covariance of the estimated model parameters for $y = X beta + epsilon $is
    given by $Var(beta) ~ (X^T X)^{-1}$.
    The determinant |Var| quantifies the volume of the confidence ellipsoid which is to
    be minimized.
    """
    eigenvalues = np.linalg.eigvalsh(X.T @ X)
    eigenvalues = eigenvalues[np.abs(eigenvalues) > delta]
    return np.sum(np.log(eigenvalues))


def a_optimality(X: np.ndarray, delta=1e-9) -> float:
    """Compute the A-optimality for a model matrix X (smaller is better).
    A-optimality is the sum of variances of the estimated model parameters, which is
    the trace of the covariance matrix $X.T @ X^-1$.

    F is symmetric positive definite, hence the trace of (X.T @ X)^-1 is equal to the
    the sum of inverse eigenvalues.
    """
    eigenvalues = np.linalg.eigvalsh(X.T @ X)
    eigenvalues = eigenvalues[np.abs(eigenvalues) > delta]
    return np.sum(1.0 / eigenvalues)  # type: ignore


def g_optimality(X: np.ndarray, delta: float = 1e-9) -> float:
    """Compute the G-optimality for a model matrix X (smaller is better).
    G-optimality is the maximum entry in the diagonal of the hat matrix
    H = X (X.T X)^-1 X.T which relates to the maximum variance of the predicted values.
    """
    H = X @ np.linalg.inv(X.T @ X + delta * np.eye(len(X))) @ X.T
    return np.max(np.diag(H))  # type: ignore


def metrics(X: np.ndarray, delta: float = 1e-9) -> pd.Series:
    """Returns a series containing D-optimality, A-optimality and G-efficiency
    for a model matrix X

    Args:
        X (np.ndarray): model matrix for which the metrics are determined
        delta (float): cutoff value for eigenvalues of the information matrix. Default value is 1e-9.

    Returns:
        A pd.Series containing the values for the three metrics.
    """
    return pd.Series(
        {
            "D-optimality": d_optimality(X, delta),
            "A-optimality": a_optimality(X, delta),
            "G-optimality": g_optimality(X, delta),
        }
    )


def check_nchoosek_constraints_as_bounds(domain: Domain) -> None:
    """Checks if NChooseK constraints of domain can be formulated as bounds.

    Args:
        domain (Domain): Domain whose NChooseK constraints should be checked
    """
    # collect NChooseK constraints
    if len(domain.constraints) == 0:
        return

    nchoosek_constraints = []
    for c in domain.constraints:
        if isinstance(c, NChooseKConstraint):
            nchoosek_constraints.append(c)

    if len(nchoosek_constraints) == 0:
        return

    # check if the domains of all NChooseK constraints are compatible to linearization
    for c in nchoosek_constraints:
        for name in np.unique(c.features):
            input = domain.inputs.get_by_key(name)
            if input.bounds[0] > 0 or input.bounds[1] < 0:  # type: ignore
                raise ValueError(
                    f"Constraint {c} cannot be formulated as bounds. 0 must be inside the \
                    domain of the affected decision variables."
                )

    # check if the parameter names of two nchoose overlap
    for c in nchoosek_constraints:
        for _c in nchoosek_constraints:
            if c != _c:
                for name in c.features:
                    if name in _c.features:
                        raise ValueError(
                            f"Domain {domain} cannot be used for formulation as bounds. \
                            names attribute of NChooseK constraints must be pairwise disjoint."
                        )


def nchoosek_constraints_as_bounds(
    domain: Domain,
    n_experiments: int,
) -> List:
    """Determines the box bounds for the decision variables

    Args:
        domain (Domain): Domain to find the bounds for.
        n_experiments (int): number of experiments for the design to be determined.

    Returns:
        A list of tuples containing bounds that respect NChooseK constraint imposed
        onto the decision variables.
    """
    check_nchoosek_constraints_as_bounds(domain)

    # bounds without NChooseK constraints
    bounds = np.array(
        [p.bounds for p in domain.inputs.get(ContinuousInput)] * n_experiments  # type: ignore
    )

    if len(domain.constraints) > 0:
        for constraint in domain.constraints:
            if isinstance(constraint, NChooseKConstraint):
                n_inactive = len(constraint.features) - constraint.max_count
                if n_inactive > 0:
                    # find indices of constraint.names in names
                    ind = [
                        i
                        for i, p in enumerate(domain.inputs.get_keys())
                        if p in constraint.features
                    ]

                    # find and shuffle all combinations of elements of ind of length max_active
                    ind = np.array(list(combinations(ind, r=n_inactive)))
                    np.random.shuffle(ind)

                    # set bounds to zero in each experiments for the variables that should be inactive
                    for i in range(n_experiments):
                        ind_vanish = ind[i % len(ind)]
                        bounds[ind_vanish + i * len(domain.inputs), :] = [0, 0]
    else:
        pass

    # convert bounds to list of tuples
    bounds = [(b[0], b[1]) for b in bounds]

    return bounds



from typing import Union

from bofire.data_models.outlier_detection.outlier_detection import (
    IterativeTrimming,
    OutlierDetection,
)
from bofire.data_models.outlier_detection.outlier_detections import (
    OutlierDetections,
)

AnyOutlierDetection = IterativeTrimming
AnyAbstractOutlierDetection = OutlierDetection



import itertools
import warnings
from typing import List, Optional

import pandas as pd
from sklearn.preprocessing import MinMaxScaler

from bofire.data_models.domain.api import Inputs
from bofire.data_models.features.api import CategoricalInput, ContinuousInput


def get_confounding_matrix(
    inputs: Inputs,
    design: pd.DataFrame,
    powers: Optional[List[int]] = None,
    interactions: Optional[List[int]] = None,
):
    """Analyzes the confounding of a design and returns the confounding matrix.

    Only takes continuous features into account.

    Args:
        inputs (Inputs): Input features.
        design (pd.DataFrame): Design matrix.
        powers (List[int], optional): List of powers of the individual factors/features that should be considered.
            Integers has to be larger than 1. Defaults to [].
        interactions (List[int], optional): List with interaction levels to be considered.
            Integers has to be larger than 1. Defaults to [2].

    Returns:
        _type_: _description_
    """
    if len(inputs.get(CategoricalInput)) > 0:
        warnings.warn("Categorical input features will be ignored.")

    keys = inputs.get_keys(ContinuousInput)
    scaler = MinMaxScaler(feature_range=(-1, 1))
    scaled_design = pd.DataFrame(
        data=scaler.fit_transform(design[keys]),
        columns=keys,
    )

    # add powers
    if powers is not None:
        for p in powers:
            assert p > 1, "Power has to be at least of degree two."
            for key in keys:
                scaled_design[f"{key}**{p}"] = scaled_design[key] ** p

    # add interactions
    if interactions is None:
        interactions = [2]

    for i in interactions:
        assert i > 1, "Interaction has to be at least of degree two."
        assert i < len(keys) + 1, f"Interaction has to be smaller than {len(keys)+1}."
        for combi in itertools.combinations(keys, i):
            scaled_design[":".join(combi)] = scaled_design[list(combi)].prod(axis=1)

    return scaled_design.corr()



from typing import Literal, Optional, Sequence

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from botorch.models.ensemble import EnsembleModel
from botorch.models.transforms.outcome import OutcomeTransform, Standardize
from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from bofire.data_models.enum import OutputFilteringEnum
from bofire.data_models.surrogates.api import MLPEnsemble as DataModel
from bofire.data_models.surrogates.scaler import ScalerEnum
from bofire.surrogates.botorch import BotorchSurrogate
from bofire.surrogates.single_task_gp import get_scaler
from bofire.surrogates.trainable import TrainableSurrogate
from bofire.utils.torch_tools import tkwargs


class RegressionDataSet(Dataset):
    """
    Prepare the dataset for regression
    """

    def __init__(self, X: Tensor, y: Tensor):
        self.X = X.to(**tkwargs)
        self.y = y.to(**tkwargs)

    def __len__(self):
        return len(self.X)

    def __getitem__(self, i: int):
        return self.X[i], self.y[i]


class MLP(nn.Module):
    def __init__(
        self,
        input_size: int,
        output_size: int = 1,
        hidden_layer_sizes: Sequence = (100,),
        dropout: float = 0.0,
        activation: Literal["relu", "logistic", "tanh"] = "relu",
    ):
        super().__init__()
        if activation == "relu":
            f_activation = nn.ReLU
        elif activation == "logistic":
            f_activation = nn.Sigmoid
        elif activation == "tanh":
            f_activation = nn.Tanh
        else:
            raise ValueError(f"Activation {activation} not known.")
        layers = [
            nn.Linear(input_size, hidden_layer_sizes[0]).to(**tkwargs),
            f_activation(),
        ]
        if dropout > 0.0:
            layers.append(nn.Dropout(dropout))
        if len(hidden_layer_sizes) > 1:
            for i in range(len(hidden_layer_sizes) - 1):
                layers += [
                    nn.Linear(hidden_layer_sizes[i], hidden_layer_sizes[i + 1]).to(
                        **tkwargs
                    ),
                    f_activation(),
                ]
                if dropout > 0.0:
                    layers.append(nn.Dropout(dropout))
        layers.append(nn.Linear(hidden_layer_sizes[-1], output_size).to(**tkwargs))
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        return self.layers(x)


class _MLPEnsemble(EnsembleModel):
    def __init__(
        self, mlps: Sequence[MLP], output_scaler: Optional[OutcomeTransform] = None
    ):
        super().__init__()
        if len(mlps) == 0:
            raise ValueError("List of mlps is empty.")
        num_in_features = mlps[0].layers[0].in_features
        num_out_features = mlps[0].layers[-1].out_features
        for mlp in mlps:
            assert mlp.layers[0].in_features == num_in_features
            assert mlp.layers[-1].out_features == num_out_features
        self.mlps = mlps
        if output_scaler is not None:
            self.outcome_transform = output_scaler
        # put all models in eval mode
        for mlp in self.mlps:
            mlp.eval()

    def forward(self, X: Tensor):
        r"""Compute the model output at X.

        Args:
            X: A `batch_shape x n x d`-dim input tensor `X`.

        Returns:
            A `batch_shape x s x n x m`-dimensional output tensor where
            `s` is the size of the ensemble.
        """
        return torch.stack([mlp(X) for mlp in self.mlps], dim=-3)

    @property
    def num_outputs(self) -> int:
        r"""The number of outputs of the model."""
        return self.mlps[0].layers[-1].out_features  # type: ignore


def fit_mlp(
    mlp: MLP,
    dataset: RegressionDataSet,
    batch_size: int = 10,
    n_epoches: int = 200,
    lr: float = 1e-4,
    shuffle: bool = True,
    weight_decay: float = 0.0,
):
    """Fit a MLP for regression to a dataset.

    Args:
        mlp (MLP): The MLP that should be fitted.
        dataset (RegressionDataSet): The data that should be fitted
        batch_size (int, optional): Batch size. Defaults to 10.
        n_epoches (int, optional): Number of training epoches. Defaults to 200.
        lr (float, optional): Initial learning rate. Defaults to 1e-4.
        shuffle (bool, optional): Whereas the batches should be shuffled. Defaults to True.
        weight_decay (float, optional): Weight decay (L2 regularization). Defaults to 0.0 (no regularization).
    """
    mlp.train()
    train_loader = DataLoader(dataset=dataset, batch_size=batch_size, shuffle=shuffle)
    loss_function = nn.L1Loss()
    optimizer = torch.optim.Adam(mlp.parameters(), lr=lr, weight_decay=weight_decay)
    for _ in range(n_epoches):
        current_loss = 0.0
        for data in train_loader:
            # Get and prepare inputs
            inputs, targets = data
            if len(targets.shape) == 1:
                targets = targets.reshape((targets.shape[0], 1))

            # Zero the gradients
            optimizer.zero_grad()

            # Perform forward pass
            outputs = mlp(inputs)

            # Compute loss
            loss = loss_function(outputs, targets)

            # Perform backward pass
            loss.backward()

            # Perform optimization
            optimizer.step()

            # Print statistics
            current_loss += loss.item()


class MLPEnsemble(BotorchSurrogate, TrainableSurrogate):
    def __init__(self, data_model: DataModel, **kwargs):
        self.n_estimators = data_model.n_estimators
        self.hidden_layer_sizes = data_model.hidden_layer_sizes
        self.activation = data_model.activation
        self.dropout = data_model.dropout
        self.batch_size = data_model.batch_size
        self.n_epochs = data_model.n_epochs
        self.lr = data_model.lr
        self.weight_decay = data_model.weight_decay
        self.subsample_fraction = data_model.subsample_fraction
        self.shuffle = data_model.shuffle
        self.scaler = data_model.scaler
        self.output_scaler = data_model.output_scaler
        super().__init__(data_model, **kwargs)

    _output_filtering: OutputFilteringEnum = OutputFilteringEnum.ALL
    model: Optional[_MLPEnsemble] = None

    def _fit(self, X: pd.DataFrame, Y: pd.DataFrame):
        scaler = get_scaler(self.inputs, self.input_preprocessing_specs, self.scaler, X)
        transformed_X = self.inputs.transform(X, self.input_preprocessing_specs)

        if self.output_scaler == ScalerEnum.STANDARDIZE:
            output_scaler = Standardize(m=Y.shape[-1])
        else:
            output_scaler = None

        mlps = []
        subsample_size = round(self.subsample_fraction * X.shape[0])
        for _ in range(self.n_estimators):
            # resample X and Y
            sample_idx = np.random.choice(X.shape[0], replace=True, size=subsample_size)
            tX = torch.from_numpy(transformed_X.values[sample_idx]).to(**tkwargs)
            ty = torch.from_numpy(Y.values[sample_idx]).to(**tkwargs)

            dataset = RegressionDataSet(
                X=scaler.transform(tX) if scaler is not None else tX,
                y=output_scaler(ty)[0] if output_scaler is not None else ty,
            )
            mlp = MLP(
                input_size=transformed_X.shape[1],
                output_size=1,
                hidden_layer_sizes=self.hidden_layer_sizes,
                activation=self.activation,  # type: ignore
                dropout=self.dropout,
            )
            fit_mlp(
                mlp=mlp,
                dataset=dataset,
                batch_size=self.batch_size,
                n_epoches=self.n_epochs,
                lr=self.lr,
                shuffle=self.shuffle,
                weight_decay=self.weight_decay,
            )
            mlps.append(mlp)
        self.model = _MLPEnsemble(mlps, output_scaler=output_scaler)
        if scaler is not None:
            self.model.input_transform = scaler



from typing import List

import gpytorch
import torch
from gpytorch.kernels import Kernel as GpytorchKernel

import bofire.data_models.kernels.api as data_models
import bofire.priors.api as priors
from bofire.kernels.fingerprint_kernels.tanimoto_kernel import TanimotoKernel


def map_RBFKernel(
    data_model: data_models.RBFKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> gpytorch.kernels.RBFKernel:
    return gpytorch.kernels.RBFKernel(
        batch_shape=batch_shape,
        ard_num_dims=len(active_dims) if data_model.ard else None,
        active_dims=active_dims,  # type: ignore
        lengthscale_prior=priors.map(data_model.lengthscale_prior)
        if data_model.lengthscale_prior is not None
        else None,
    )


def map_MaternKernel(
    data_model: data_models.MaternKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> gpytorch.kernels.MaternKernel:
    return gpytorch.kernels.MaternKernel(
        batch_shape=batch_shape,
        ard_num_dims=len(active_dims) if data_model.ard else None,
        active_dims=active_dims,
        nu=data_model.nu,
        lengthscale_prior=priors.map(data_model.lengthscale_prior)
        if data_model.lengthscale_prior is not None
        else None,
    )


def map_LinearKernel(
    data_model: data_models.LinearKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> gpytorch.kernels.LinearKernel:
    return gpytorch.kernels.LinearKernel(
        batch_shape=batch_shape,
        active_dims=active_dims,
        variance_prior=priors.map(data_model.variance_prior)
        if data_model.variance_prior is not None
        else None,
    )


def map_PolynomialKernel(
    data_model: data_models.PolynomialKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> gpytorch.kernels.PolynomialKernel:
    return gpytorch.kernels.PolynomialKernel(
        batch_shape=batch_shape,
        active_dims=active_dims,
        power=data_model.power,
        offset_prior=priors.map(data_model.offset_prior)
        if data_model.offset_prior is not None
        else None,
    )


def map_AdditiveKernel(
    data_model: data_models.AdditiveKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> gpytorch.kernels.AdditiveKernel:
    return gpytorch.kernels.AdditiveKernel(
        *[  # type: ignore
            map(
                data_model=k,
                batch_shape=batch_shape,
                ard_num_dims=ard_num_dims,
                active_dims=active_dims,
            )
            for k in data_model.kernels
        ]
    )


def map_MultiplicativeKernel(
    data_model: data_models.MultiplicativeKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> gpytorch.kernels.ProductKernel:
    return gpytorch.kernels.ProductKernel(
        *[  # type: ignore
            map(
                data_model=k,
                batch_shape=batch_shape,
                ard_num_dims=ard_num_dims,
                active_dims=active_dims,
            )
            for k in data_model.kernels
        ]
    )


def map_ScaleKernel(
    data_model: data_models.ScaleKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> gpytorch.kernels.ScaleKernel:
    return gpytorch.kernels.ScaleKernel(
        base_kernel=map(
            data_model.base_kernel,
            batch_shape=batch_shape,
            ard_num_dims=ard_num_dims,
            active_dims=active_dims,
        ),
        outputscale_prior=priors.map(data_model.outputscale_prior)
        if data_model.outputscale_prior is not None
        else None,
    )


def map_TanimotoKernel(
    data_model: data_models.TanimotoKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> TanimotoKernel:
    return TanimotoKernel(
        batch_shape=batch_shape,
        ard_num_dims=len(active_dims) if data_model.ard else None,
        active_dims=active_dims,
    )


KERNEL_MAP = {
    data_models.RBFKernel: map_RBFKernel,
    data_models.MaternKernel: map_MaternKernel,
    data_models.LinearKernel: map_LinearKernel,
    data_models.PolynomialKernel: map_PolynomialKernel,
    data_models.AdditiveKernel: map_AdditiveKernel,
    data_models.MultiplicativeKernel: map_MultiplicativeKernel,
    data_models.ScaleKernel: map_ScaleKernel,
    data_models.TanimotoKernel: map_TanimotoKernel,
}


def map(
    data_model: data_models.AnyKernel,
    batch_shape: torch.Size,
    ard_num_dims: int,
    active_dims: List[int],
) -> GpytorchKernel:
    return KERNEL_MAP[data_model.__class__](  # type: ignore
        data_model, batch_shape, ard_num_dims, active_dims
    )



import uuid
from contextlib import contextmanager
from pathlib import Path
from tempfile import gettempdir
from typing import Optional


@contextmanager
def make_tmpfile(name: Optional[str] = None):
    name = name or uuid.uuid4().hex
    test_folder = Path(gettempdir()) / "bofire"
    test_folder.mkdir(parents=True, exist_ok=True)
    test_file = test_folder / name
    try:
        yield test_file
    finally:
        test_file.unlink()



from typing import List

import mock
import numpy as np
import pandas as pd
import pytest
from _pytest.fixtures import fixture
from pandas.testing import assert_frame_equal
from pydantic.error_wrappers import ValidationError

import tests.bofire.strategies.dummy as dummy
from bofire.data_models.constraints.api import (
    LinearEqualityConstraint,
    LinearInequalityConstraint,
    NChooseKConstraint,
)
from bofire.data_models.domain.api import Domain, Outputs
from bofire.data_models.features.api import (
    CategoricalInput,
    ContinuousInput,
    ContinuousOutput,
)
from bofire.data_models.objectives.api import TargetObjective
from bofire.data_models.outlier_detection.api import OutlierDetections
from bofire.data_models.outlier_detection.outlier_detection import IterativeTrimming
from bofire.data_models.surrogates.api import SingleTaskGPSurrogate
from bofire.strategies.strategy import Strategy
from tests.bofire.data_models.test_domain_validators import (
    generate_candidates,
    generate_experiments,
)
from tests.bofire.strategies.specs import (
    VALID_CATEGORICAL_INPUT_FEATURE_SPEC,
    VALID_CONTINUOUS_INPUT_FEATURE_SPEC,
    VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
    VALID_LINEAR_EQUALITY_CONSTRAINT_SPEC,
    VALID_LINEAR_INEQUALITY_CONSTRAINT_SPEC,
    VALID_NCHOOSEKE_CONSTRAINT_SPEC,
)

if1 = ContinuousInput(
    **{**VALID_CONTINUOUS_INPUT_FEATURE_SPEC, "key": "if1", "bounds": (0, 5.3)}
)
if2 = ContinuousInput(
    **{**VALID_CONTINUOUS_INPUT_FEATURE_SPEC, "key": "if2", "bounds": (0, 5.3)}
)
if3 = CategoricalInput(
    **{
        **VALID_CATEGORICAL_INPUT_FEATURE_SPEC,
        "key": "if3",
    }
)

of1 = ContinuousOutput(
    **{
        **VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
        "key": "of1",
    }  # type: ignore
)
of2 = ContinuousOutput(
    **{
        **VALID_CONTINUOUS_OUTPUT_FEATURE_SPEC,
        "key": "of2",
    }  # type: ignore
)
of3 = ContinuousOutput(key="of3", objective=None)

of4 = ContinuousOutput(
    key="of4",
    objective=TargetObjective(w=1, target_value=5.0, tolerance=1.0, steepness=0.5),
)

c1 = LinearEqualityConstraint(
    **{
        **VALID_LINEAR_EQUALITY_CONSTRAINT_SPEC,
        "features": ["if1", "if2"],
        "coefficients": [1, 1],
    }
)
c2 = LinearInequalityConstraint(
    **{
        **VALID_LINEAR_INEQUALITY_CONSTRAINT_SPEC,
        "features": ["if1", "if2"],
        "coefficients": [1, 1],
    }
)
c3 = NChooseKConstraint(
    **{
        **VALID_NCHOOSEKE_CONSTRAINT_SPEC,
        "features": ["if1", "if2"],
    }
)


@fixture  # type: ignore
def strategy():
    data_model = dummy.DummyStrategyDataModel(
        domain=Domain.from_lists(
            inputs=[if1, if2],
            outputs=[of1, of2],
            constraints=[],
        )
    )
    return dummy.DummyStrategy(data_model=data_model)


@pytest.mark.parametrize(
    "domain",
    [
        (
            Domain.from_lists(
                inputs=[if1, if2],
                outputs=[of1],
                constraints=constraints,
            )
        )
        for constraints in [[c1], [c2], [c1, c2]]
    ],
)
def test_strategy_constructor(
    domain: Domain,
):
    dummy.DummyStrategyDataModel(domain=domain)


@pytest.mark.parametrize(
    "domain",
    [
        (
            Domain.from_lists(
                inputs=[if1, if2],
                outputs=[of1],
                constraints=constraints,
            )
        )
        for constraints in [[c3], [c1, c3], [c2, c3], [c1, c2, c3]]
    ],
)
def test_strategy_init_domain_invalid_constraints(
    domain: Domain,
):
    with pytest.raises(ValidationError):
        dummy.DummyStrategyDataModel(domain=domain)


@pytest.mark.parametrize(
    "domain",
    [
        (
            Domain.from_lists(
                inputs=inputs,
                outputs=[of1],
                constraints=[],
            )
        )
        for inputs in [[if3], [if1, if3]]
    ],
)
def test_strategy_init_domain_invalid_input(domain: Domain):
    with pytest.raises(ValidationError):
        dummy.DummyStrategyDataModel(domain=domain)


@pytest.mark.parametrize(
    "domain",
    [
        (
            Domain.from_lists(
                inputs=[if1, if2],
                outputs=outputs,
                constraints=[],
            )
        )
        for outputs in [[of1, of4], [of4]]
    ],
)
def test_strategy_init_domain_invalid_objective(domain: Domain):
    with pytest.raises(ValidationError):
        dummy.DummyStrategyDataModel(domain=domain)


def test_strategy_init_domain_noobjective():
    domain = Domain.from_lists(
        inputs=[if1, if2],
        outputs=[of3],
        constraints=[],
    )
    with pytest.raises(ValidationError):
        dummy.DummyStrategyDataModel(domain=domain)


domain = Domain.from_lists(
    inputs=[if1, if2],
    outputs=[of1, of2],
    constraints=[],
)
e1 = generate_experiments(domain, 1)
e2 = generate_experiments(domain, 2)
e3 = generate_experiments(domain, 3)
e4 = generate_experiments(domain, 4)


@pytest.mark.parametrize(
    "domain, experiments, replace",
    [
        (
            Domain.from_lists(
                inputs=[if1, if2],
                outputs=[of1, of2],
                constraints=[],
            ),
            experiments,
            replace,
        )
        for experiments in [e1, e2]
        for replace in [True, False]
    ],
)
def test_strategy_tell_initial(
    domain: Domain,
    experiments: pd.DataFrame,
    replace: bool,
):
    """verify that tell correctly stores initial experiments"""
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    strategy.tell(experiments=experiments, replace=replace)
    assert strategy.experiments.equals(experiments)


def test_strategy_no_variance():
    domain = Domain.from_lists(
        inputs=[
            ContinuousInput(
                key="a",
                bounds=(0, 1),
            ),
            ContinuousInput(key="b", bounds=(1, 1)),
        ],
        outputs=[of1],
    )
    experiments = domain.inputs.sample(5)
    experiments["of1"] = [1, 2, 3, 4, 5]
    experiments["valid_of1"] = 1
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    with pytest.raises(ValueError):
        strategy.tell(experiments)
    # introduce variance but in an invalid experiment
    experiments.loc[0, "valid_of1"] = 0
    experiments.loc[0, "b"] = 0.7
    with pytest.raises(ValueError):
        strategy.tell(experiments)
    # introduce variance
    experiments["b"] = 0
    strategy.tell(experiments)


def test_strategy_set_experiments():
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    assert strategy.num_experiments == 0
    experiments = generate_experiments(domain, 2)
    strategy.set_experiments(experiments=experiments)
    assert_frame_equal(strategy.experiments, experiments)
    assert_frame_equal(strategy._experiments, experiments)  # type: ignore
    assert strategy.num_experiments == 2


def test_strategy_add_experiments():
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    assert strategy.num_experiments == 0
    experiments = generate_experiments(domain, 2)
    strategy.add_experiments(experiments=experiments)
    assert_frame_equal(strategy.experiments, experiments)
    assert strategy.num_experiments == 2
    experiments2 = generate_experiments(domain, 5)
    strategy.add_experiments(experiments=experiments2)
    assert strategy.num_experiments == 7
    assert_frame_equal(
        strategy.experiments, pd.concat((experiments, experiments2), ignore_index=True)
    )


def test_strategy_set_candidates():
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    assert strategy.num_candidates == 0
    candidates = generate_candidates(domain, 2)
    strategy.set_candidates(candidates=candidates)
    assert_frame_equal(strategy.candidates, candidates[domain.inputs.get_keys()])
    assert_frame_equal(strategy._candidates, candidates[domain.inputs.get_keys()])  # type: ignore
    assert strategy.num_candidates == 2
    strategy.reset_candidates()
    assert strategy.num_candidates == 0


def test_strategy_add_candidates():
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    assert strategy.num_candidates == 0
    candidates = generate_candidates(domain, 2)
    strategy.add_candidates(candidates=candidates)
    assert_frame_equal(strategy.candidates, candidates[domain.inputs.get_keys()])
    assert strategy.num_candidates == 2
    candidates2 = generate_candidates(domain, 5)
    strategy.add_candidates(candidates=candidates2)
    assert strategy.num_candidates == 7
    assert_frame_equal(
        strategy.candidates,
        pd.concat((candidates, candidates2), ignore_index=True)[
            domain.inputs.get_keys()
        ],
    )


@pytest.mark.parametrize(
    "domain, experimentss",
    [
        (
            Domain.from_lists(
                inputs=[if1, if2],
                outputs=[of1, of2],
                constraints=[],
            ),
            experimentss,
        )
        for experimentss in [[e1, e2], [e2, e1], [e1, e2, e3, e4]]
    ],
)
def test_strategy_tell_append(
    domain: Domain,
    experimentss: List[pd.DataFrame],
):
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    for index, experiments in enumerate(experimentss):
        strategy.tell(experiments=experiments, replace=False)
        expected_len = sum([len(e) for e in experimentss[: index + 1]])
        assert len(strategy.experiments) == expected_len


@pytest.mark.parametrize(
    "domain, experimentss",
    [(domain, experimentss) for experimentss in [[e1, e2], [e2, e1], [e1, e2, e3, e4]]],
)
def test_strategy_tell_replace(
    domain: Domain,
    experimentss: List[pd.DataFrame],
):
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    for experiments in experimentss:
        strategy.tell(experiments=experiments, replace=True)
        expected_len = len(experiments)
        assert len(strategy.experiments) == expected_len


@pytest.mark.parametrize(
    "domain",
    [domain],
)
def test_strategy_tell_outliers(
    domain: Domain,
):
    experiments = generate_experiments(domain=domain, row_count=200)
    outlier_detectors = []
    for i, key in enumerate(domain.outputs.get_keys()):
        experiments.loc[:59, key] = experiments.loc[:59, key] + np.random.randn(60) * 1
        outlier_detectors.append(
            IterativeTrimming(
                base_gp=SingleTaskGPSurrogate(
                    inputs=domain.inputs, outputs=Outputs(features=[domain.outputs[i]])
                )
            )
        )
    experiments = domain.validate_experiments(experiments=experiments)
    experiments1 = experiments.copy()
    strategy = dummy.DummyBotorchPredictiveStrategy(
        data_model=dummy.DummyStrategyDataModel(
            domain=domain,
            outlier_detection_specs=OutlierDetections(detectors=outlier_detectors),
        )
    )
    strategy1 = dummy.DummyBotorchPredictiveStrategy(
        data_model=dummy.DummyStrategyDataModel(
            domain=domain,
        )
    )
    strategy.tell(experiments=experiments)
    assert_frame_equal(
        experiments1, experiments
    )  # test that experiments don't get changed outside detect_outliers
    strategy1.tell(experiments=experiments)
    assert str(strategy.model.state_dict()) != str(strategy1.model.state_dict())  # type: ignore # test if two fitted surrogates are different


@pytest.mark.parametrize("domain, experiments", [(domain, e) for e in [e3, e4]])
def test_strategy_ask_invalid_candidates(
    domain: Domain,
    experiments: pd.DataFrame,
):
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    strategy.tell(experiments)

    def test_ask(self: Strategy, candidate_count: int):
        candidates = generate_candidates(self.domain, candidate_count)
        candidates = candidates.drop(candidates.columns[0], axis=1)
        return candidates

    with mock.patch.object(dummy.DummyStrategy, "_ask", new=test_ask):
        with pytest.raises(ValueError):
            strategy.ask(candidate_count=1)


@pytest.mark.parametrize("domain, experiments", [(domain, e) for e in [e3, e4]])
def test_strategy_ask_invalid_candidate_count(
    domain: Domain,
    experiments: pd.DataFrame,
):
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    strategy.tell(experiments)

    def test_ask(self: Strategy, candidate_count: int):
        candidates = generate_candidates(self.domain, candidate_count)[:-1]
        return candidates

    with mock.patch.object(dummy.DummyStrategy, "_ask", new=test_ask):
        with pytest.raises(ValueError):
            strategy.ask(candidate_count=4)


@pytest.mark.parametrize("domain, experiments", [(domain, e) for e in [e3, e4]])
def test_strategy_ask_valid(
    domain: Domain,
    experiments: pd.DataFrame,
):
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    strategy.tell(experiments)

    def test_ask(self: Strategy, candidate_count: int):
        candidates = generate_candidates(self.domain, candidate_count)
        return candidates

    with mock.patch.object(dummy.DummyStrategy, "_ask", new=test_ask):
        strategy.ask(candidate_count=1)


def test_ask_invalid_candidate_count_request():
    strategy = dummy.DummyStrategy(
        data_model=dummy.DummyStrategyDataModel(domain=domain)
    )
    strategy.tell(e3)
    with pytest.raises(ValueError):
        strategy.ask(-1)


# test for PredictiveStrategy
@pytest.mark.parametrize(
    "domain, experiments",
    [
        (
            Domain.from_lists(
                inputs=[if1, if2],
                outputs=[of1, of2],
                constraints=[],
            ),
            e,
        )
        for e in [e3, e4]
    ],
)
def test_predictive_strategy_ask_valid(
    domain: Domain,
    experiments: pd.DataFrame,
):
    strategy = dummy.DummyPredictiveStrategy(
        data_model=dummy.DummyPredictiveStrategyDataModel(domain=domain)
    )
    strategy.tell(experiments)

    def test_ask(self: Strategy, candidate_count: int):
        candidates = generate_candidates(self.domain, candidate_count)
        return candidates

    with mock.patch.object(dummy.DummyPredictiveStrategy, "_ask", new=test_ask):
        strategy.ask(candidate_count=1)


def test_predictivestrategy_to_candidates():
    domain = Domain.from_lists(
        inputs=[if1, if2],
        outputs=[of1, of2],
        constraints=[],
    )
    strategy = dummy.DummyPredictiveStrategy(
        data_model=dummy.DummyPredictiveStrategyDataModel(domain=domain)
    )
    candidates = generate_candidates(domain, 5)
    strategy.to_candidates(candidates=candidates)


def test_predictive_strategy_ask_invalid():
    """Test that PretictiveStrategy also checks if candidates contain output columns."""
    strategy = dummy.DummyPredictiveStrategy(
        data_model=dummy.DummyPredictiveStrategyDataModel(
            domain=Domain.from_lists(
                inputs=[if1, if2],
                outputs=[of1, of2],
                constraints=[],
            )
        )
    )
    strategy.tell(e3)

    def test_ask(self: Strategy, candidate_count: int):
        candidates = generate_candidates(self.domain, candidate_count)
        return candidates.drop(columns=["of1_pred"])

    with mock.patch.object(dummy.DummyPredictiveStrategy, "_ask", new=test_ask):
        with pytest.raises(ValueError):
            strategy.ask(candidate_count=2)


@pytest.mark.parametrize(
    "domain, experiments",
    [
        (
            Domain.from_lists(
                inputs=[if1, if2],
                outputs=[of1, of2],
                constraints=[],
            ),
            e,
        )
        for e in [e3, e4]
    ],
)
def test_predictive_strategy_predict(domain, experiments):
    strategy = dummy.DummyPredictiveStrategy(
        data_model=dummy.DummyPredictiveStrategyDataModel(domain=domain)
    )
    strategy.tell(experiments)
    preds = strategy.predict(generate_candidates(domain=domain))
    assert sorted(preds.columns) == sorted(
        [
            "of1_pred",
            "of2_pred",
            "of1_sd",
            "of2_sd",
            "of1_des",
            "of2_des",
        ]
    )


@pytest.mark.parametrize(
    "domain",
    [
        (
            Domain.from_lists(
                inputs=[if1, if2],
                outputs=[of1, of2],
                constraints=[],
            )
        )
    ],
)
def test_predictive_strategy_predict_not_fitted(domain):
    strategy = dummy.DummyPredictiveStrategy(
        data_model=dummy.DummyPredictiveStrategyDataModel(domain=domain)
    )
    with pytest.raises(ValueError):
        strategy.predict(generate_candidates(domain=domain))



from typing import Union

from bofire.data_models.acquisition_functions.acquisition_function import (
    AcquisitionFunction,
    MultiObjectiveAcquisitionFunction,
    SingleObjectiveAcquisitionFunction,
    qEHVI,
    qEI,
    qLogEHVI,
    qLogEI,
    qLogNEHVI,
    qLogNEI,
    qNEHVI,
    qNEI,
    qPI,
    qSR,
    qUCB,
)

AbstractAcquisitionFunction = [
    AcquisitionFunction,
    SingleObjectiveAcquisitionFunction,
    MultiObjectiveAcquisitionFunction,
]

AnyAcquisitionFunction = Union[
    qNEI, qEI, qSR, qUCB, qPI, qLogEI, qLogNEI, qEHVI, qLogEHVI, qNEHVI, qLogNEHVI
]

AnySingleObjectiveAcquisitionFunction = Union[
    qNEI, qEI, qSR, qUCB, qPI, qLogEI, qLogNEI
]

AnyMultiObjectiveAcquisitionFunction = Union[qEHVI, qLogEHVI, qNEHVI, qLogNEHVI]
